"""
Module to check if all training/testing scenarios have the same goal-configuration and automatically configure
the goal_observations in configs.yaml
"""

__author__ = "Armin Ettenhofer"
__copyright__ = "TUM Cyber-Physical System Group"
__credits__ = [""]
__version__ = "1.0"
__maintainer__ = "Armin Ettenhofer"
__email__ = "armin.ettenhofer@tum.de"
__status__ = "Released"

import argparse
import logging
import os
import pickle
import re
from typing import List, Set

from commonroad.scenario.trajectory import State

import commonroad_rl.gym_commonroad.constants as constants

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
LOGGER.addHandler(handler)


def get_args():
    """Scan arguments"""
    parser = argparse.ArgumentParser(
        description="Analyzes goal_definitions of scenarios and configures goal_observations for model",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--problem_dir", "-i", help="Path to problems", type=str,
                        default=os.path.join(constants.PATH_PARAMS["pickles"], "problem"))
    parser.add_argument("--configure", "-c", help="Adjust the goal_observations in the configuration",
                        action="store_true")
    parser.add_argument("--config_path", "-cp", help="Path to model configuration file", type=str,
                        default=constants.PATH_PARAMS["configs"]["commonroad-v1"])
    return parser.parse_args()


def main(problem_dir: str = os.path.join(constants.PATH_PARAMS["pickles"], "problem"), configure: bool = False,
         config_path: str = constants.PATH_PARAMS["configs"]["commonroad-v1"]) -> bool:
    """
    Analyzes goal-requirements of scenarios and sets goal-observations in config

    :param problem_dir: Path to problem directory
    :param configure: If the goal_observations should be configured
    :param config_path: Path to the environment configuration file
    :return: If all scenarios have the same goal definition
    """

    # Check arguments
    assert os.path.isdir(problem_dir), "The problem directory doesn't exist"
    if configure:
        assert os.path.isfile(config_path), "The path to the config file is invalid!"

    # Find all pickled scenarios
    _, _, scenarios_pickled = next(os.walk(problem_dir))
    scenarios_pickled = [filename for filename in scenarios_pickled if os.path.splitext(filename)[1] == ".pickle"]
    if len(scenarios_pickled) == 0:
        LOGGER.warning("The problem directory does not contain any problems!")
        return True

    # Analyze the goal_states
    goal_states = analyze_problems(base_path=problem_dir, filenames=scenarios_pickled)

    # Evaluate the results
    matching: bool = goal_states.pop("matching")
    position: bool = goal_states["position"]
    velocity: bool = goal_states["velocity"]
    orientation: bool = goal_states["orientation"]
    time_step: bool = goal_states["time_step"]

    if matching:
        print("The goal states of all problems have the same attributes.")
    else:
        print("Not all goal states of the problems match.\n")
    print("The recommended goal_observation configuration is:")
    print(f"Position: {position}, Velocity: {velocity}, Orientation: {orientation}, Time_step: {time_step}\n")

    # Set config.yaml if desired
    if not configure:
        return matching

    set_configuration(config_path=config_path, goal_observation=goal_states)
    return matching


def analyze_problems(base_path: str, filenames: [str]) -> dict:
    """
    Checks if the goal states of all problems have the same definitions and returns the resulting definition

    :param base_path: Path to problem directory
    :param filenames: Names of all the problems to be analyzed
    :return: Goal definition
    """
    print(f"Analyzing {len(filenames)} files.\n")

    init = False
    matching = True
    attributes: Set[str] = set()

    # Look at every goal state in every planning problem in every scenario
    for filename in filenames:
        with open(os.path.join(base_path, filename), 'rb') as file:
            planning_problem_dict: dict = pickle.load(file)["planning_problem_set"].planning_problem_dict

        for _, planning_problem in planning_problem_dict.items():
            goal_state_list: List[State] = planning_problem.goal.state_list
            for goal_state in goal_state_list:
                if init:
                    goal_attributes = set(goal_state.attributes)
                    if goal_attributes != attributes:
                        # Print the first not matching file
                        if matching:
                            print(f"\'{os.path.splitext(filename)[0]}\' and \'{os.path.splitext(filenames[0])[0]}\'"
                                  f" don't have the same goal_state attributes.\n")
                        matching = False
                        attributes |= goal_attributes

                # Initialize with the first goal_state
                else:
                    init = True
                    attributes = set(goal_state.attributes)

    config = {"position": False, "velocity": False, "orientation": False, "time_step": False, "matching": matching}
    for attribute in attributes:
        config[attribute] = True

    return config


def set_configuration(config_path: str, goal_observation: dict):
    """
    Sets the goal_observations in the config according to the input goal_observation

    :param config_path: Path to the config-file (.yaml file)
    :param goal_observation: How the goal observations should be configured
    """
    print("Setting the goal_observations in the configs.yaml\n")
    with open(config_path, "r") as file:
        config = file.read()

    if goal_observation["position"]:
        config = re.sub("observe_distance_goal_long:( ){0,5}([Tt]rue|[Ff]alse)",
                        "observe_distance_goal_long: True", config)
        config = re.sub("observe_distance_goal_lat:( ){0,5}([Tt]rue|[Ff]alse)",
                        "observe_distance_goal_lat: True", config)
    else:
        config = re.sub("observe_distance_goal_long:( ){0,5}([Tt]rue|[Ff]alse)",
                        "observe_distance_goal_long: False", config)
        config = re.sub("observe_distance_goal_lat:( ){0,5}([Tt]rue|[Ff]alse)",
                        "observe_distance_goal_lat: False", config)

    if goal_observation["velocity"]:
        config = re.sub("observe_distance_goal_velocity:( ){0,5}([Tt]rue|[Ff]alse)",
                        "observe_distance_goal_velocity: True", config)
    else:
        config = re.sub("observe_distance_goal_velocity:( ){0,5}([Tt]rue|[Ff]alse)",
                        "observe_distance_goal_velocity: False", config)

    if goal_observation["orientation"]:
        config = re.sub("observe_distance_goal_orientation:( ){0,5}([Tt]rue|[Ff]alse)",
                        "observe_distance_goal_orientation: True", config)
    else:
        config = re.sub("observe_distance_goal_orientation:( ){0,5}([Tt]rue|[Ff]alse)",
                        "observe_distance_goal_orientation: False", config)

    if goal_observation["time_step"]:
        config = re.sub("observe_distance_goal_time:( ){0,5}([Tt]rue|[Ff]alse)",
                        "observe_distance_goal_time: True", config)
    else:
        config = re.sub("observe_distance_goal_time:( ){0,5}([Tt]rue|[Ff]alse)",
                        "observe_distance_goal_time: False", config)

    with open(config_path, "w") as file:
        file.write(config)


if __name__ == "__main__":
    args = get_args()
    main(
        problem_dir=args.problem_dir,
        configure=args.configure,
        config_path=args.config_path
    )



__author__ = "Peter Kocsis"
__copyright__ = "TUM Cyber-Physical System Group"
__credits__ = []
__version__ = "0.1"
__maintainer__ = "Peter Kocsis"
__email__ = "peter.kocsis@tum.de"
__status__ = "Integration"

"""
Integration tests of the CommonRoad-RL repository
"""
import logging
import os
import shutil

from commonroad_rl.train_model import run_stable_baselines, run_stable_baselines_argsparser
from commonroad_rl.generate_solution import solve_scenarios
from commonroad_rl.tests.common.marker import *
from commonroad_rl.tests.common.path import resource_root, output_root
from commonroad_rl.tools.pickle_scenario.xml_to_pickle import pickle_xml_scenarios

logging.root.setLevel(logging.DEBUG)

resource_path = resource_root("test_commonroad_rl")
output_path = output_root("test_commonroad_rl")


def run_overfit(test_batch, goal_relaxation, num_of_steps, env_id):
    xml_scenarios_path = os.path.join(resource_path, test_batch)
    output_base_path = os.path.join(output_path, test_batch)
    pickle_path = os.path.join(output_base_path, "pickles")
    log_path = os.path.join(output_base_path, "logs")
    solution_path = os.path.join(output_base_path, "solutions")

    # Pickle CommonRoad scenarios
    pickle_xml_scenarios(
        input_dir=xml_scenarios_path,
        output_dir=pickle_path
    )

    # Overfit model
    meta_scenario_path = os.path.join(pickle_path, "meta_scenario")
    train_reset_config_path = os.path.join(pickle_path, "problem")
    test_reset_config_path = os.path.join(pickle_path, "problem")
    shutil.copytree(test_reset_config_path, os.path.join(pickle_path, "problem_test"))
    visualization_path = os.path.join(output_path, "images")
    algo = "ppo2"

    args_str = (
        f"--algo {algo} --env {env_id} --seed 1 --eval-freq 1000 --log-folder {log_path} --n-timesteps {num_of_steps}"
        f" --info_keywords is_collision is_time_out is_off_road is_goal_reached --env-kwargs"
        f' reward_type:"hybrid_reward"'
        f' meta_scenario_path:"{meta_scenario_path}"'
        f' train_reset_config_path:"{train_reset_config_path}"'
        f' test_reset_config_path:"{test_reset_config_path}"'
        f' visualization_path:"{visualization_path}" '
    )

    if env_id == "commonroad-v1":
        args_str += "goal_configs:{'relax_is_goal_reached':" + f"{goal_relaxation}" + "} "
        # TODO: force other necessary settings for this scenario
        args_str += "surrounding_configs:{'observe_lane_circ_surrounding':" + f"{True}" + "}"
        args_str += " surrounding_configs:{'observe_lidar_circle_surrounding':" + f"{False}" + "}"
        args_str += " reward_configs_hybrid:{'reward_get_close_coefficient':2.," \
                    "'reward_goal_reached':1000.,'reward_collision':-1000.}"
        args_str += " vehicle_params:{'vehicle_type':2,'vehicle_model':0}"

    args = run_stable_baselines_argsparser().parse_args(args_str.split(sep=" "))
    run_stable_baselines(args)  # , save_path=os.path.join(output_path, "model"))

    # Solve scenarios
    model_path = os.path.join(log_path, algo, f"{env_id}_1")
    cost_function = "JB1"

    results = solve_scenarios(
        test_path=pickle_path,
        model_path=model_path,
        algo=algo,
        solution_path=solution_path,
        cost_function=cost_function,
        env_id=env_id,
        render=False, model_name="commonroad-v1"
    )

    assert all(results), f"not all overfit scenarios solved: {results}"


@pytest.mark.parametrize(
    ("env_id", "test_batch", "goal_relaxation", "num_of_steps"),
    [("commonroad-v1", "DEU_A9-2_1_T-1", False, 6000)]
)
@functional
@integration_test
def test_overfit_model(env_id, test_batch, goal_relaxation, num_of_steps):
    run_overfit(test_batch, goal_relaxation, num_of_steps, env_id)


@pytest.mark.parametrize(
    ("env_id", "test_batch", "goal_relaxation", "num_of_steps"),
    [("commonroad-v1", "DEU_A99-1_1_T-1", False, 30000)],
)
@slow
@functional
@integration_test
def test_overfit_model_slow(env_id, test_batch, goal_relaxation, num_of_steps):
    run_overfit(test_batch, goal_relaxation, num_of_steps, env_id)



from commonroad_rl.gym_commonroad.observation.observation import Observation
from commonroad_rl.gym_commonroad.observation.ego_observation import EgoObservation
from commonroad_rl.gym_commonroad.observation.goal_observation import GoalObservation
from commonroad_rl.gym_commonroad.observation.surrounding_observation import SurroundingObservation
from commonroad_rl.gym_commonroad.observation.lanelet_network_observation import LaneletNetworkObservation
from commonroad_rl.gym_commonroad.observation.traffic_sign_observation import TrafficSignObservation
from commonroad_rl.gym_commonroad.observation.observation_collector import ObservationCollector



"""
Module for scenario related helper methods for the CommonRoad Gym environment
"""
import numpy as np
from typing import Tuple, Union, List, Set
from commonroad.common.solution import VehicleModel, VehicleType
from commonroad.prediction.prediction import TrajectoryPrediction
from commonroad.scenario.lanelet import Lanelet
from commonroad.scenario.obstacle import DynamicObstacle, StaticObstacle
from commonroad.scenario.trajectory import State
from commonroad.scenario.scenario import ScenarioID, Scenario
from commonroad_dc.feasibility.feasibility_checker import trajectory_feasibility
from commonroad_dc.feasibility.vehicle_dynamics import VehicleDynamics, VehicleParameters
from shapely.geometry import LineString


def parse_map_name(scenario_id: ScenarioID) -> str:
    return f"{scenario_id.map_name}-{str(scenario_id.map_id)}"


def make_valid_orientation(angle: float) -> float:
    # make orientation [-pi, pi]
    # TODO: update this function in commonroad.common.util
    TWO_PI = 2.0 * np.pi
    angle = angle % TWO_PI
    if np.pi <= angle <= TWO_PI:
        angle = angle - TWO_PI
    assert -np.pi <= angle <= np.pi
    return angle


def get_road_edge(scenario) -> Tuple[dict, dict, dict, dict]:
    """
    TODO only used in preprocessing -> relocate there maybe??

    Get the road edge or solid white line of a lanelet.
    :return: Dictionary of left and right lanelet ids and road edge
    """
    left_road_edge_lanelet_id = {}
    right_road_edge_lanelet_id = {}
    left_road_edge = {}
    right_road_edge = {}
    for lanelet in scenario.lanelet_network.lanelets:
        if lanelet.lanelet_id not in right_road_edge_lanelet_id.keys():
            # Search the right most lanelet and use right vertices as right bound
            start = lanelet
            temp = []
            while start.adj_right_same_direction:
                temp.append(start.lanelet_id)
                start = scenario.lanelet_network.find_lanelet_by_id(start.adj_right)
            temp.append(start.lanelet_id)
            right_bound = LineString(start.right_vertices)
            right_road_edge_lanelet_id.update({k: start.lanelet_id for k in temp})
            right_road_edge[start.lanelet_id] = right_bound
        if lanelet.lanelet_id not in left_road_edge_lanelet_id.keys():
            start = lanelet
            temp = []
            # Search the left most lanelet and use right vertices as left bound
            while start.adj_left_same_direction:
                temp.append(start.lanelet_id)
                start = scenario.lanelet_network.find_lanelet_by_id(start.adj_left)
            temp.append(start.lanelet_id)
            left_bound = LineString(start.left_vertices)
            left_road_edge_lanelet_id.update({k: start.lanelet_id for k in temp})
            left_road_edge[start.lanelet_id] = left_bound
    return (
        left_road_edge_lanelet_id,
        left_road_edge,
        right_road_edge_lanelet_id,
        right_road_edge,
    )


def get_lane_marker(ego_vehicle_lanelet: Lanelet) -> Tuple[LineString, LineString]:
    """
    TODO remove, only used in old commonroad_env class

    Get the lane marker for the desired lanelet.

    :param ego_vehicle_lanelet: lanelet of ego vehicle
    :return: left and right lane marker
    """
    left_marker_line = LineString(ego_vehicle_lanelet.left_vertices)
    right_marker_line = LineString(ego_vehicle_lanelet.right_vertices)
    return left_marker_line, right_marker_line


def interpolate_steering_angles(
        state_list: List[State], parameters: VehicleParameters, dt: float
) -> List[State]:
    """
    Interpolates the not defined steering angles based on KS Model

    :param state_list: The list of the states
    :param parameters: The parameters of the vehicle
    :param dt: dt of the scenario
    :return: The state list with interpolated steering angles
    """
    if len(state_list) == 0:
        return state_list

    l_wb = parameters.a + parameters.b

    [orientations, velocities] = np.array(
        [[state.orientation, state.velocity] for state in state_list]
    ).T

    orientation_vectors = approx_orientation_vector(orientations)
    psi_dots = (
            angle_difference(orientation_vectors[:, :-1].T, orientation_vectors[:, 1:].T) / dt
    )
    avg_velocities = np.mean(np.array([velocities[:-1], velocities[1:]]), axis=0)
    avg_velocities[avg_velocities == 0.0] += np.finfo(float).eps

    steering_angles = np.arctan(psi_dots * l_wb / avg_velocities)
    if len(steering_angles) > 0:
        steering_angles = np.hstack((steering_angles, steering_angles[-1]))
    else:
        default_steering_angle = 0.0
        steering_angles = np.array([default_steering_angle])

    steering_angles = np.clip(
        steering_angles, parameters.steering.min, parameters.steering.max
    )

    def get_state_with_steering_angle(state: State, steering_angle: float):
        if hasattr(state, "steering_angle"):
            if state.steering_angle is None:
                state.steering_angle = steering_angle
        else:
            setattr(state, "steering_angle", steering_angle)
        return state

    return list(
        map(
            lambda state, steering_angle: get_state_with_steering_angle(
                state, steering_angle
            ),
            state_list,
            steering_angles,
        )
    )


def interpolate_steering_angles_of_obstacle(obstacle: DynamicObstacle, parameters: VehicleParameters, dt: float):
    """
    Interpolates the not defined steering angles of obstacle based on KS Model

    :param obstacle:
    :param parameters: The parameters of the vehicle
    :param dt: dt of the scenario
    """
    trajectory = obstacle.prediction.trajectory
    trajectory.state_list = interpolate_steering_angles(trajectory.state_list, parameters, dt)
    obstacle.initial_state.steering_angle = trajectory.state_list[0].steering_angle


def check_trajectory(
        obstacle: DynamicObstacle,
        vehicle_model: VehicleModel,
        vehicle_type: VehicleType,
        dt: float,
) -> bool:
    """
    Checks whether the trajectory of a given obstacle is feasible with a given vehicle model
    Note: Currently it is implemented for the KS model. As soon as the workaround is not needed,
    it can be rebased to fully use the Feasibility Checker

    :param obstacle: The obstacle which trajectory should be checked
    :param vehicle_model: The used vehicle model
    :param vehicle_type: THe type of the vehicle
    :param dt: Delta time of the simulation
    :return: True if the trajectory is feasible
    """
    trajectory = obstacle.prediction.trajectory
    vehicle_dynamics = VehicleDynamics.from_model(vehicle_model, vehicle_type)

    position_tolerance = 0.1
    orientation_tolerance = 2e-2

    e = np.array([position_tolerance, position_tolerance, orientation_tolerance])

    feasible, _ = trajectory_feasibility(trajectory, vehicle_dynamics, dt, e=e)
    return feasible


def approx_orientation_vector(orientation: Union[float, np.ndarray]) -> np.ndarray:
    """
    Approximate normed vector in a given orientation

    :param orientation: The orientation
    :return Normalized vector points to the defined orientation
    """
    return np.array([np.cos(orientation), np.sin(orientation)])


def angle_difference(vector_from: np.ndarray, vector_to: np.ndarray):
    """
    Returns angle between the two provided vectors, from v1 to v2

    :param vector_from: Vector from the angle should be measured
    :param vector_to: Vector to the angle should be measured
    :return: Signed relative angle between the two vectors
    """
    assert vector_from.ndim <= 2 and vector_to.ndim <= 2
    if vector_from.ndim == 1:
        vector_from = vector_from[None]
    if vector_to.ndim == 1:
        vector_to = vector_to[None]
    dot_product = np.einsum("ij,ij->i", vector_from, vector_to)
    determinant = np.einsum("ij,ij->i", vector_from, (np.flip(vector_to, axis=-1) * (np.array([1, -1]))))

    return np.arctan2(determinant, dot_product)


"""
Navigation connecting the RoutePlanner with Observations
"""
from typing import List, Union, Tuple, Set
from enum import Enum
from scipy.spatial import KDTree
import numpy as np
import math
import logging
from commonroad.planning.goal import GoalRegion
from commonroad.scenario.lanelet import Lanelet
from commonroad.scenario.scenario import Scenario
from commonroad.scenario.trajectory import State
from commonroad_route_planner.route_planner import RouteType
from commonroad_route_planner.route import Route
import commonroad_dc.pycrccosy as pycrccosy
from commonroad_dc.geometry.util import resample_polyline, compute_polyline_length
import commonroad.geometry.shape as cr_shape
from shapely.geos import LOG
from shapely.ops import unary_union
from shapely.geometry import Polygon

LOGGER = logging.getLogger(__name__)


class Navigator:
    class CosyVehicleObservation(Enum):
        """Enum for the observations CoSy"""
        # TODO: add this as a property for the class, not for every function call.

        AUTOMATIC = "automatic"
        LOCALCARTESIAN = "local_cartesian"
        VEHICLEFRAME = "vehicle_frame_cosy"

        @classmethod
        def values(cls):
            return [item.value for item in cls]

    def __init__(self, route: Route):
        # version 2020 # version 06.2021 variables shared
        assert route is not None
        self.scenario = route.scenario
        self.lanelet_network = self.scenario.lanelet_network
        self.planning_problem = route.planning_problem
        self.route = route

        # version 2020 variables
        self.sectionized_environment = self.route.retrieve_route_sections()
        self.sectionized_environment_set = set(
            [item for sublist in self.sectionized_environment for item in sublist]
        )
        self.merged_route_lanelets = None
        self.ccosy_list = self._get_route_cosy()

        self.num_of_lane_changes = len(self.ccosy_list)
        self.merged_section_length_list = np.array(
            [curvi_cosy.length() for curvi_cosy in self.ccosy_list]
        )
        # ==================== #
        #         Goal         #
        # ==================== #

        self._initialize_goal()

        # version 06.2021 variables
        # ==================== #
        #    Curvy Lane Dict #
        # ==================== #
        self.lane_ccosy_kd_ref_dict = {}
        self._initialize_lane_ccosy_kd_ref_dict()
        self._initialized_goal_ref_path()

    def _initialize_goal(self):
        # version 2020
        self.goal_curvi_face_coords = None

        if self.route.type == RouteType.REGULAR:
            goal_face_coords = self._get_goal_face_points(
                self._get_goal_polygon(self.planning_problem.goal)
            )
            self.goal_curvi_face_coords = np.array(
                [
                    (self._get_safe_curvilinear_coords(self.ccosy_list[-1], g))[0]
                    for g in goal_face_coords
                ]
            )

            self.goal_min_curvi_coords = np.min(self.goal_curvi_face_coords, axis=0)
            self.goal_max_curvi_coord = np.max(self.goal_curvi_face_coords, axis=0)

    def _initialized_goal_ref_path(self):
        """
        computes the longituginal coordinates of the goal on the route.reference path
        """

        # version 06.2021
        # default for Survival Scenarios: Goal is everywhere reached.
        self.goal_min_curvi_coords_ref = -np.inf
        self.goal_max_curvi_coord_ref = np.inf

        if self.route.type == RouteType.REGULAR:
            # for Regular Scenarios: Goal is everywhere reached.
            goal_face_coords = self._get_goal_face_points(
                self._get_goal_polygon(self.planning_problem.goal)
            )
            self.goal_curvi_face_coords_ref = np.array(
                [
                    (
                        self._get_safe_distance_to_curvilinear2(
                            id_curvilinear="reference_path", position_ego=pos_g
                        )
                    )[0]
                    for pos_g in goal_face_coords
                ]
            )
            self.goal_min_curvi_coords_ref = np.amin(self.goal_curvi_face_coords_ref)
            self.goal_max_curvi_coord_ref = np.amax(self.goal_curvi_face_coords_ref)

    def _initialize_lane_ccosy_kd_ref_dict(self):
        """
        initializes the lane_ccosy_kd_ref_dict with Curvilinearsystems and kdtree
        """
        # version 06.2021
        id_polyline = [("reference_path", self.route.reference_path)]
        for id_lanelet in self.route.list_ids_lanelets:
            id_polyline.append(
                (
                    id_lanelet,
                    self.route.scenario.lanelet_network.find_lanelet_by_id(
                        id_lanelet
                    ).center_vertices,
                )
            )

        for id_curvilinear, polyline in id_polyline:

            ccosy, resampled_polyline = self.create_coordinate_system_from_polyline(
                polyline, return_resampled_polyline=True
            )

            self.lane_ccosy_kd_ref_dict[id_curvilinear] = (
                ccosy,
                KDTree(resampled_polyline),
                resampled_polyline,
            )

    def _get_route_cosy(
        self,
    ) -> Union[pycrccosy.CurvilinearCoordinateSystem, List[Lanelet]]:
        """
        merges the lanelets of reference path from start to goal to successing lanelets
        """
        # version 2020
        self.merged_route_lanelets = []

        # Append predecessor of the initial to ensure that the goal state is not out of the projection domain
        # initial_lanelet = self.lanelet_network.find_lanelet_by_id(self.route.route[0])
        # predecessors_lanelet = initial_lanelet.predecessor
        # if predecessors_lanelet is not None and len(predecessors_lanelet) != 0:
        #     predecessor_lanelet = self.lanelet_network.find_lanelet_by_id(predecessors_lanelet[0])
        #     current_merged_lanelet = predecessor_lanelet
        # else:
        #     current_merged_lanelet = None
        current_merged_lanelet = None

        for current_lanelet_id, next_lanelet_id in zip(
            self.route.list_ids_lanelets[:-1], self.route.list_ids_lanelets[1:]
        ):
            lanelet = self.lanelet_network.find_lanelet_by_id(current_lanelet_id)
            # If the lanelet is the end of a section, then change section
            if next_lanelet_id not in lanelet.successor:
                if current_merged_lanelet is not None:
                    self.merged_route_lanelets.append(current_merged_lanelet)
                    current_merged_lanelet = None
            else:
                if current_merged_lanelet is None:
                    current_merged_lanelet = lanelet
                else:
                    current_merged_lanelet = Lanelet.merge_lanelets(
                        current_merged_lanelet, lanelet
                    )

        goal_lanelet = self.lanelet_network.find_lanelet_by_id(
            self.route.list_ids_lanelets[-1]
        )
        if current_merged_lanelet is not None:
            current_merged_lanelet = Lanelet.merge_lanelets(
                current_merged_lanelet, goal_lanelet
            )
        else:
            current_merged_lanelet = goal_lanelet

        # Append successor of the goal to ensure that the goal state is not out of the projection domain
        # goal_lanelet = self.lanelet_network.find_lanelet_by_id(self.route.route[-1])
        # successors_of_goal = goal_lanelet.successor
        # if successors_of_goal is not None and len(successors_of_goal) != 0:
        #     successor_lanelet = self.lanelet_network.find_lanelet_by_id(successors_of_goal[0])
        #     current_merged_lanelet = Lanelet.merge_lanelets(current_merged_lanelet, successor_lanelet)

        self.merged_route_lanelets.append(current_merged_lanelet)

        return [
            self.create_coordinate_system_from_polyline(merged_lanelet.center_vertices)
            for merged_lanelet in self.merged_route_lanelets
        ]

    @classmethod
    def create_coordinate_system_from_polyline(
        cls,
        polyline: np.ndarray,
        return_resampled_polyline: bool = False,
    ) -> Tuple[pycrccosy.CurvilinearCoordinateSystem, np.ndarray]:
        """
        create CurvilinearCoordinateSystem from resampled polyline.

        :param polyline: np.ndarray[float[int>1, 2]] polyline
        :return:
            CurvilinearCoordinateSystem
            Resampled Polyline if flag return_resampled_polyline is set.
        """
        # version 2020, # version 06.2021
        assert polyline.ndim == 2

        if polyline.ndim == 2 and polyline.shape[0] == 2:
            step = 0.5
        else:
            step = min(compute_polyline_length(polyline) / 10.0, 0.5)
        resampled_polyline = resample_polyline(polyline, step=step)

        # remove consecutive duplicates, because they hurt runtime and may cause out of domain errors
        resampled_polyline_shifted = np.roll(resampled_polyline, 1, axis=0)
        # slice all x,y points, where the next point is np.isclose to previous one
        index_slice = np.any(np.invert(np.isclose(resampled_polyline, resampled_polyline_shifted)), axis=1)
        # sort out duplicated points
        resampled_polyline = resampled_polyline[index_slice]

        if resampled_polyline.shape[0] <= 3:
            LOGGER.warning(
                f"resampled polyline for CurvilinearCoSy has less then three points."
                f"all resampled points are  on the edge of the domain. resampled_polyline: {resampled_polyline} polyline {polyline}"
            )
        ccosy = pycrccosy.CurvilinearCoordinateSystem(resampled_polyline)
        if return_resampled_polyline:
            return ccosy, resampled_polyline
        else:
            return ccosy

    def _get_safe_curvilinear_coords(
        self, ccosy, position: np.ndarray
    ) -> Tuple[np.ndarray, int]:
        # version 2020

        try:
            rel_pos_to_domain = 0
            long_lat_distance = self._get_curvilinear_coords(ccosy, position)
        except ValueError:
            long_lat_distance, rel_pos_to_domain = self._project_out_of_domain(
                ccosy, position
            )

        return np.array(long_lat_distance), rel_pos_to_domain

    def _project_out_of_domain(
        self, ccosy, position: np.ndarray
    ) -> Tuple[np.ndarray, int]:
        """for projection out of domain.
        Extend start and end of ccosy linearily with tangent"""
        # version 2020

        eps = 0.0001
        curvi_coords_of_projection_domain = np.array(
            ccosy.curvilinear_projection_domain()
        )

        longitudinal_min, normal_min = (
            np.min(curvi_coords_of_projection_domain, axis=0) + eps
        )
        longitudinal_max, normal_max = (
            np.max(curvi_coords_of_projection_domain, axis=0) - eps
        )
        normal_center = (normal_min + normal_max) / 2
        bounding_points = np.array(
            [
                ccosy.convert_to_cartesian_coords(longitudinal_min, normal_center),
                ccosy.convert_to_cartesian_coords(longitudinal_max, normal_center),
            ]
        )
        rel_positions = position - np.array(
            [bounding_point for bounding_point in bounding_points]
        )
        distances = np.linalg.norm(rel_positions, axis=1)

        if distances[0] < distances[1]:
            # Nearer to the first bounding point
            rel_pos_to_domain = -1
            long_dist = longitudinal_min + np.dot(
                ccosy.tangent(longitudinal_min), rel_positions[0]
            )
            lat_dist = normal_center + np.dot(
                ccosy.normal(longitudinal_min), rel_positions[0]
            )
        else:
            # Nearer to the last bounding point
            rel_pos_to_domain = 1
            long_dist = longitudinal_max + np.dot(
                ccosy.tangent(longitudinal_max), rel_positions[1]
            )
            lat_dist = normal_center + np.dot(
                ccosy.normal(longitudinal_max), rel_positions[1]
            )

        return np.array([long_dist, lat_dist]), rel_pos_to_domain

    def _get_curvilinear_coords(self, ccosy, position: np.ndarray) -> np.ndarray:
        # version 06.2021
        return ccosy.convert_to_curvilinear_coords(position[0], position[1])

    def _get_curvilinear_coords_over_lanelet(self, lanelet: Lanelet, position):
        # version 2020
        current_ccosy = self.create_coordinate_system_from_polyline(
            lanelet.center_vertices
        )

        return self._get_curvilinear_coords(current_ccosy, position)

    @staticmethod
    def _get_goal_face_points(goal_shape: Polygon):
        """
        Extracts the middle points of each face of the goal region
        NOTE in pathological examples, this can still result in points outside the coordinate system
        however for the points on both ends of the lanelet, they coincide with the center vertices,
        which is what the curvilinear coordinate system is based on
        NOTE if the goal areas edges are not all intersecting with any lanelet in the ccosy, this operation will fail
        :param goal_shape: shape of the goal area
        :return: tuples of x,y coordinates of the middle points of each face of the goal region
        """
        # version 2020 # version 06.2021
        assert isinstance(goal_shape, Polygon), (
            f"Only single Polygon is supported, but {type(goal_shape)} was given,"
            f" Use a planning problem with contiguous goal region"
        )

        goal_coords = [np.array(x) for x in zip(*goal_shape.exterior.coords.xy)]

        # round the same precision as is done within the commonroad xml files
        goal_coords = [
            np.round((a + b) / 2, 6) for a, b in zip(goal_coords, goal_coords[1:])
        ]
        return goal_coords

    def _get_goal_polygon(self, goal: GoalRegion) -> Polygon:
        """
        Get the goal position as Polygon
        :param goal: the goal given as a GoalRegion
        :return: Polygon of the goal position
        """
        # version 2020 # version 06.2021

        def get_polygon_list_from_shapegroup(
            shapegroup: cr_shape.ShapeGroup,
        ) -> List[Polygon]:
            """
            Converts cr_shape.ShapeGroup to list of Polygons
            :param shapegroup: the ShapeGroup to be converted
            :return: The list of the polygons
            """

            polygon_list = []
            for shape in shapegroup.shapes:
                if isinstance(shape, cr_shape.ShapeGroup):
                    polygon_list.append(get_polygon_list_from_shapegroup(shape))
                elif isinstance(shape, (cr_shape.Rectangle, cr_shape.Polygon)):
                    polygon_list.append(shape.shapely_object)
                else:
                    raise ValueError(
                        f"Shape can't be converted to Shapely Polygon: {shape}"
                    )
            return polygon_list

        def merge_polygons(polygons_to_merge):
            return unary_union(
                [
                    geom if geom.is_valid else geom.buffer(0)
                    for geom in polygons_to_merge
                ]
            )

        polygons = [Polygon()]
        for goal_state in goal.state_list:
            if hasattr(goal_state, "position"):
                if isinstance(goal_state.position, cr_shape.ShapeGroup):
                    polygons.extend(
                        get_polygon_list_from_shapegroup(goal_state.position)
                    )
                elif isinstance(
                    goal_state.position, (cr_shape.Rectangle, cr_shape.Polygon)
                ):
                    polygons.append(goal_state.position.shapely_object)
                else:
                    raise NotImplementedError(
                        f"Goal position not supported yet, "
                        f"only ShapeGroup, Rectangle or Polygon shapes can be used, "
                        f"the given shape was: {type(goal_state.position)}"
                    )

        merged_polygon = merge_polygons(polygons)
        return merged_polygon

    def get_position_curvi_coords(self, ego_vehicle_state_position: np.ndarray):
        # version 2020
        for cosy_idx, curvi_cosy in enumerate(self.ccosy_list):

            ego_curvi_coords, rel_pos_to_domain = self._get_safe_curvilinear_coords(
                curvi_cosy, ego_vehicle_state_position
            )

            is_last_section = cosy_idx == self.num_of_lane_changes - 1
            if rel_pos_to_domain == 1 and not is_last_section:
                continue

            return ego_curvi_coords, cosy_idx

        raise ValueError("Unable to project the ego vehicle on the global cosy")

    def get_long_lat_distance_to_goal(
        self, ego_vehicle_state_position: np.ndarray
    ) -> Tuple[float, float]:
        """
        Get the longitudinal and latitudinal distance from the ego vehicle to the goal,
        measured in the frenet cos of the Navigator lanelet
        :param ego_vehicle_state_position: position of the ego vehicle
        :return:
            longitudinal distance of closest point on ref_path to goal
            latitudinal distance if in domain,
        """
        # version 2020
        # If the route is survival, then return zero
        if self.route.type == RouteType.SURVIVAL:
            return 0.0, 0.0

        ego_curvi_coords, cosy_idx = self.get_position_curvi_coords(
            ego_vehicle_state_position
        )

        is_last_section = cosy_idx == self.num_of_lane_changes - 1

        if is_last_section:
            relative_distances = self.goal_curvi_face_coords - ego_curvi_coords
            min_distance = np.min(relative_distances, axis=0)
            max_distance = np.max(relative_distances, axis=0)

            (min_distance_long, min_distance_lat) = np.maximum(
                np.minimum(0.0, max_distance), min_distance
            )
        else:
            min_distance_long = (
                self.merged_section_length_list[cosy_idx] - ego_curvi_coords[0]
            )
            current_section_idx = cosy_idx + 1
            while current_section_idx != self.num_of_lane_changes - 1:
                min_distance_long += self.merged_section_length_list[
                    current_section_idx
                ]
                current_section_idx += 1

            relative_lat_distances = (
                self.goal_curvi_face_coords[:, 1] - ego_curvi_coords[1]
            )
            min_distance = np.min(relative_lat_distances, axis=0)
            max_distance = np.max(relative_lat_distances, axis=0)

            min_distance_long += self.goal_min_curvi_coords[0]
            min_distance_lat = np.maximum(np.minimum(0.0, max_distance), min_distance)

        return min_distance_long, min_distance_lat

    def _get_safe_distance_to_curvilinear2(
        self, id_curvilinear: Union[str, int], position_ego: np.ndarray
    ):
        """
        distance to curvilinear with fallback to closest point on reference path via kdtree
        :param id_curvilinear: unique string, lanelet_id or "reference_path"
        :param position_ego: np.ndarray[float[2, 1]],

        :return:
            p_curvilinear_closest[0]: float[0, ccosy.length()], closest longitudinal coordinate to position_ego
            distance: float[0, ccosy.length()], distance to closest point on reference path
            ccosy: pycrccosy.CurvilinearCoordinateSystem
            indomain: float, 0.0, if position_ego in domain, else 1
        """
        # version 06.2021
        if id_curvilinear not in self.lane_ccosy_kd_ref_dict:
            # create if this is the first time of inferencing this lanelet_id
            raise BaseException(
                f" error in _get_safe_distance_to_curvilinear2: id_curvilinear not in self.lane_ccosy_kd_ref_dict {id_curvilinear} {self.lane_ccosy_kd_ref_dict.keys()}"
            )

        ccosy, kdtree, resampled_polyline = self.lane_ccosy_kd_ref_dict[id_curvilinear]

        indomain = 1.0
        try:
            p_curvilinear_closest = ccosy.convert_to_curvilinear_coords(
                position_ego[0], position_ego[1]
            )
            distance = p_curvilinear_closest[1]
            indomain = 1.0

        except ValueError:
            try:
                # for projection out of domain
                # LOGGER.debug(
                #     f"projection of ego vehicle {position_ego} to reference path of id {id_curvilinear} CCoSy is out of domain. "
                #     "defaulting to closest distance on reference path."
                # )
                # get index of 2D reference_path array to which position_ego is closest
                distance, idx = kdtree.query(position_ego)
                closest_point = resampled_polyline[idx]

                p_curvilinear_closest = ccosy.convert_to_curvilinear_coords(
                    closest_point[0], closest_point[1]
                )
                # Done, but distance is yet not in curvilinear direction.
                # return indomain as angle between ego and point on ccosy to the curvilinear CoSy.
                if np.isclose(distance, 0, atol=1e-3):
                    indomain = 1
                else:
                    tangent = self.safe_tangent(ccosy, p_curvilinear_closest[0])
                    tangent = tangent / np.linalg.norm(tangent)
                    to_closest = position_ego - closest_point
                    # projected on the tangent of the closest curvilinear, take portion which is normal to
                    lateral_distance_to_tangent = np.cross(tangent, to_closest)
                    # add sign to the distance,
                    distance = math.copysign(distance, lateral_distance_to_tangent)

                    # indomain is now cos of the angle of the ego towards point w. tangent on ccosy
                    indomain = lateral_distance_to_tangent / distance
            except Exception as ex:
                raise BaseException(
                    f" error in _get_safe_distance_to_curvilinear2: failed in id_curvilinear {id_curvilinear} closest_point {closest_point} idx{idx} resampled_polyline {resampled_polyline}. \n Exception: {ex}"
                )
        except Exception as ex:
            raise BaseException(
                f" error in _get_safe_distance_to_curvilinear2: failed in id_curvilinear {id_curvilinear} resampled_polyline {resampled_polyline}. \n Exception: {ex}"
            )

        return p_curvilinear_closest[0], distance, ccosy, indomain

    def get_lane_change_distance(
        self, state: State, active_lanelets: List[int] = None
    ) -> float:
        # version 2020
        """
        get distance in the frenet cos of the reference path to the end of the current lanelet


        """
        # If the route is survival, then return zero
        if self.route.type == RouteType.SURVIVAL:
            return 0.0
        if active_lanelets is None:
            active_lanelets = self.scenario.lanelet_network.find_lanelet_by_position(
                [state.position]
            )[0]

        current_lanelet_ids_on_route = [
            current_lanelet_id
            for current_lanelet_id in active_lanelets
            if current_lanelet_id in self.sectionized_environment_set
        ]
        # The state is not on the route, instant lane change is required
        if len(current_lanelet_ids_on_route) == 0:
            return 0.0

        sorted_current_lanelet_ids_on_route = Navigator.sorted_lanelet_ids(
            current_lanelet_ids_on_route,
            state.orientation,
            state.position,
            self.scenario,
        )

        # The most likely current lanelet id by considering the orientation of the state
        current_lanelet_id = sorted_current_lanelet_ids_on_route[0]

        distance_until_lane_change = 0.0
        route_successors = {current_lanelet_id}
        while len(route_successors) != 0:
            # Add the length of the current lane
            current_lanelet_id = route_successors.pop()
            current_lanelet = self.lanelet_network.find_lanelet_by_id(
                current_lanelet_id
            )
            try:
                if distance_until_lane_change == 0.0:
                    # Calculate the remaining distance in this lanelet
                    current_distance = self._get_curvilinear_coords_over_lanelet(
                        current_lanelet, state.position
                    )
                    current_distance_long = current_distance[0]
                    distance_until_lane_change = (
                        current_lanelet.distance[-1] - current_distance_long
                    )
                else:
                    distance_until_lane_change += current_lanelet.distance[-1]

            except ValueError:
                pass

            successors_set = set(current_lanelet.successor)
            route_successors = successors_set.intersection(
                self.sectionized_environment_set
            )

        return distance_until_lane_change

    @classmethod
    def _evaluate_orientation_observations(
        cls,
        state=State,
        observation_cos: CosyVehicleObservation = CosyVehicleObservation.AUTOMATIC,
    ) -> float:
        """get orientation of the ego cos for observations.

        CosyVehicleObservation.VEHICLEFRAME:
            always rotated in ego orienation (fallback velocity heading)
        CosyVehicleObservation.LOCALCARTESIAN:
            never rotated / orienation in global east /
        CosyVehicleObservation.AUTOMATIC:
            VEHICLEFRAME if state has orientation, else LOCALCARTESIAN

        """
        # version 06.2021
        if observation_cos is None:
            observation_cos = cls.CosyVehicleObservation.AUTOMATIC

        if observation_cos == cls.CosyVehicleObservation.LOCALCARTESIAN:
            # orienation not considered
            return 0.0
        elif hasattr(state, "orientation") and observation_cos in [
            cls.CosyVehicleObservation.AUTOMATIC,
            cls.CosyVehicleObservation.VEHICLEFRAME,
        ]:
            return float(state.orientation)
        else:
            # for PM model get orientation from velocity
            if observation_cos == cls.CosyVehicleObservation.VEHICLEFRAME:
                # get orientation from velocity
                if hasattr(state, "velocity_y") and hasattr(state, "velocity"):
                    return np.arctan2(state.velocity_y, state.velocity)
                else:
                    raise BaseException(
                        "State has no orienation or global x & y orienations. Cant compute vehicle frame CoSy."
                    )
            else:
                return 0.0

    def get_waypoints_of_reference_path(
        self,
        state,
        distances_ref_path: Set[float],
        observation_cos: CosyVehicleObservation = CosyVehicleObservation.AUTOMATIC,
    ):
        """
        computes the vectors in observation_cos to the waypoints in the reference path ccosy.
        waypoints are on the reference path, at lonitudinal positions: distances_ref_path,
        compared to the closest position of the ego on the reference path
        returning the vectors to the points on reference path in ego local_cosy and relative orientaion.

        :param distances_ref_path: Set of Waypoints at distance d to be computed. Include int(0)
        :param state: state of the ego vehicle
        :param observation_cos: string, in which vehicle coordinate system observation should be made
        :return:
            np.ndarray: vectors to points, shape: (len(distances_ref_path), 2)
            np.ndarray: orientations at points, shape: (len(distances_ref_path), )

        minimal example:
        Navigator.get_waypoints_of_reference_path(
            state=EgoVehicle.State,
            distances_ref_path=[-1000, 0, 1000],
            observation_cos=CosyVehicleObservation.AUTOMATIC,
        )
        """
        # version 06.2021
        position_ego = state.position

        orientation_vehicle = self._evaluate_orientation_observations(
            state, observation_cos
        )

        # self.route.clcs is pycrccosy CurvilinearCoordinateSystem of route.ref_path
        # position closet on reference path in clcs: p_curvilinear = [long_clcs, lat_clcs]
        p_curvilinear, _, ccosy, _ = self._get_safe_distance_to_curvilinear2(
            id_curvilinear="reference_path", position_ego=position_ego
        )
        vectors_to_ref, orientations_ref = self.get_points_on_ccosy(
            ccosy, p_curvilinear, distances_ref_path, orientation_vehicle, position_ego
        )
        return (vectors_to_ref, orientations_ref)

    def get_longlat_togoal_on_reference_path(self, state) -> float:
        """
        returns the distance  [m] to the closest point on the reference path,
        and the distance to the end of the curvilinear cosy

        :param state: state of ego
        :return:
            float, longitudinal distance to point of reference path which is closest to goal
            float, distance to reference path
            float, indomain:
                1 if indomain (distance to reference path orthogonal to ccosy),
                0-1 if angle towards domain not orthodgonal

        minimal example:
        float, float = get_longlat_togoal_on_reference_path(np.array([0,10]))
        """
        # version 06.2021

        (
            p_curvilinear_long,
            distance,
            ccosy,
            indomain,
        ) = self._get_safe_distance_to_curvilinear2(
            id_curvilinear="reference_path", position_ego=state.position
        )

        if p_curvilinear_long < self.goal_min_curvi_coords_ref:
            min_distance_to_goal_long_on_ref = (
                self.goal_min_curvi_coords_ref - p_curvilinear_long
            )
        elif p_curvilinear_long > self.goal_max_curvi_coord_ref:
            min_distance_to_goal_long_on_ref = (
                self.goal_max_curvi_coord_ref - p_curvilinear_long
            )
        else:
            min_distance_to_goal_long_on_ref = 0.0

        return min_distance_to_goal_long_on_ref, distance, indomain

    @staticmethod
    def safe_convert_to_cartesian_coords(
        ccosy: pycrccosy.CurvilinearCoordinateSystem,
        long_lat_eval: Tuple[float, float],
        eps: float = 0.001,
        max_lat: float = 20.0,
        length_ccosy: Union[float, None] = None,
    ) -> Tuple[float, float]:
        """evaluate from ccosy to cartesian with respect to boundaries

        :param ccosy: Curvilinear System
        :param long_lat_eval: long and lat position in ccosy, for which cartesian shall be returned
        :optional:
            :param eps: nearest longitudinal distance allowed to ends of ccosy. default: 0.001m
            :param max_lat: nearest lateral distance allowed to center of ccosy.
                default: 20m == default max set during ccosy creation
            :param length_ccosy: length of ccosy, if held in memory

        :return: corresponding cartesian position
        """
        try:
            if length_ccosy is None:
                length_ccosy = ccosy.length()
            long_clcs_eval = min(max(long_lat_eval[0], eps), length_ccosy - eps)
            lat_clcs_eval = min(max(long_lat_eval[1], -max_lat + eps), max_lat)
            return ccosy.convert_to_cartesian_coords(long_clcs_eval, lat_clcs_eval)
        except ValueError as ex:
            raise BaseException(
                f"try to evaluate long: {long_clcs_eval} lat: {lat_clcs_eval} on ccosy but got ValueError {ex}"
            )

    @staticmethod
    def safe_tangent(
        ccosy: pycrccosy.CurvilinearCoordinateSystem,
        long_eval: float,
        eps: float = 0.001,
        length_ccosy: Union[float, None] = None,
    ) -> Tuple[float, float]:
        """evaluate from tanget at a position in the ccosy with respect to boundaries

        :param ccosy: Curvilinear System
        :param long_eval: long  position in ccosy, for which tangent shall be returned
        :optional:
            :param eps: nearest longitudinal distance allowed to ends of ccosy. default: 0.001m
            :param length_ccosy:  length of ccosy, if held in memory, default: None

        :return: corresponding cartesian position
        """
        if length_ccosy is None:
            length_ccosy = ccosy.length()
        long_clcs_eval = min(max(long_eval, eps), length_ccosy - eps)
        return ccosy.tangent(long_clcs_eval)

    @classmethod
    def get_points_on_ccosy(
        cls,
        ccosy: pycrccosy.CurvilinearCoordinateSystem,
        p_curvilinear: float,
        distances_ref_path: set,
        orientation_ego: float,
        position_ego: np.ndarray,
    ):
        """
        computes the vectors in ego/local_cosy to the points for (p_curvilinear+distances_ref_path) in ccosy.
        returning the vectors to the points on reference path in ego local_cosy and relative orientaion.

        :raises ValueError: if (p_curvilinear, 0) not in ccosy domain

        :param distances_ref_path: Set of Waypoints at distance d to be computed. Include int(0)
        :param state: state of the ego vehicle
        :return:
            np.ndarray: vectors to points, shape: (len(distances_ref_path), 2)
            np.ndarray: orientations at points, shape: (len(distances_ref_path), )

        minimal example:
        positions, lane_orienations = Navigator.get_points_on_ccosy(
            ccosy = pycrccosy.CurvilinearCoordinateSystem(),
            p_curvilinear = 5.1,
            distances_ref_path = [-np.inf, -5, 0, 10, 50],
            orientation_ego = np.pi,
            position_ego = np.array([0, 0])
        )
        """
        # version 06.2021
        vectors_to_ref = []
        v_ref_tangent = []

        distances_ref_path = sorted(distances_ref_path)

        length_ccosy = ccosy.length()

        projected_point = None
        projected_tangent = None

        # get observations: cartesian distance and orientation of reference waypoints in vehicle_cos
        for distance_ref in distances_ref_path:
            # distance in positive direction of the clcs
            long_clcs_eval = p_curvilinear + distance_ref
            # make valid in ccosy

            try:
                # try to get true distance
                projected_point = cls.safe_convert_to_cartesian_coords(
                    ccosy,
                    long_lat_eval=[long_clcs_eval, 0.0],
                    length_ccosy=length_ccosy,
                )
                projected_tangent = cls.safe_tangent(ccosy, long_clcs_eval)
            except ValueError as ex:
                LOGGER.error(
                    f"coordinate at long {long_clcs_eval} {length_ccosy} has exception {ex}. padding with last value: {projected_point}"
                )
                # use padding with last working value.
                if projected_point is None:
                    projected_point = cls.safe_convert_to_cartesian_coords(
                        ccosy,
                        long_lat_eval=[p_curvilinear, 0.0],
                        length_ccosy=length_ccosy,
                    )
                    projected_tangent = cls.safe_tangent(ccosy, p_curvilinear)
            finally:
                vectors_to_ref.append(projected_point)
                v_ref_tangent.append(projected_tangent)

        # positions and orientations of waypoints collected, for lower variance project now on vehicle cos
        vectors_to_ref = np.asarray(vectors_to_ref) - position_ego  # relative vector
        v_ref_tangent = np.asarray(
            v_ref_tangent
        )  # unit length tangent vector at dist i of ref_path
        # compute relative orientations at the ref_path points
        orientations_ref = (
            np.arctan2(v_ref_tangent[:, 1], v_ref_tangent[:, 0]) - orientation_ego
        )
        # rotate matrix arond orientation, so
        vectors_to_ref, _ = cls._rotate_2darray_by_orientation(
            orient=-orientation_ego, twodim_array=vectors_to_ref
        )
        return vectors_to_ref, orientations_ref

    @staticmethod
    def _rotate_2darray_by_orientation(
        orient: float,
        twodim_array: np.ndarray,
        rot_matrix: Union[None, np.matrix] = None,
    ) -> np.ndarray:
        """
        rotates "list of 2d poins" array by orientation

        :param orient: float [-pi, pi]
        :param twodim_array: np.ndarray[float[int, 2]]
        :param rot_matrix: None or np.matrix[2, 2]
        :return:
            np.ndarray: rotated twodim_array shape (n,2)
            np.matrix: np.ndarray shape (2,2)
        """

        # version 06.2021
        assert (
            twodim_array.shape[0] > 0 and twodim_array.shape[1] == 2
        ), f"imput twodim_array has wrong shape {twodim_array.shape}"
        if float(orient) == 0.0:
            return twodim_array, np.asarray([[1, 0], [0, 1]])
        elif rot_matrix is None:
            c = np.cos(orient)
            s = np.sin(orient)
            rot_matrix = np.array([[c, s], [-s, c]]).T
        else:
            assert rot_matrix.shape == (2, 2)

        # return rotated array and rotation matrix for further usage
        return (rot_matrix @ twodim_array.T).T, rot_matrix

    def get_referencepath_multilanelets_waypoints(
        self,
        state: State,
        distances_per_lanelet: Set[float],
        lanelets_id_rel: Set[int] = (-1, 0, 1, 2),
        observation_cos: CosyVehicleObservation = CosyVehicleObservation.AUTOMATIC,
    ):
        """
        returns the vectors and orientations for waypoints on a set of selected lanelets (centerlines).
        include lanelets which are at lanelets_id_rel relative to current closest lanelet to ego.
        (0: current, -1 predecessor in planned route, +1 next in planned route etc.)
        compute waypoints at positions: distances_per_lanelet for each lanelet.

        :param state: ego state
        :lanelets_id_rel: List[int]
        lanelets_id_rel: List[int] = [0, 1]
        """
        # version 06.2021

        computed_waypoints = np.zeros(
            (len(lanelets_id_rel), len(distances_per_lanelet), 2)
        )
        computed_orientations = np.zeros(
            (len(lanelets_id_rel), len(distances_per_lanelet))
        )

        position_ego = state.position

        orientation_vehicle = self._evaluate_orientation_observations(
            state, observation_cos
        )

        distance_lanelets = []
        # self.route.clcs is pycrccosy CurvilinearCoordinateSystem of route.ref_path
        # position closet on reference path in clcs: p_curvilinear = [long_clcs, lat_clcs]
        for id_lanelet in self.route.list_ids_lanelets:
            if id_lanelet not in self.lane_ccosy_kd_ref_dict:
                raise BaseException(
                    f" error in get_referencepath_multilanelets_waypoints: id_lanelet not in self.lane_ccosy_kd_ref_dict {id_lanelet} {self.lane_ccosy_kd_ref_dict.keys()}"
                )

            (
                p_curvilinear_long,
                distance,
                ccosy,
                _,
            ) = self._get_safe_distance_to_curvilinear2(
                id_curvilinear=id_lanelet, position_ego=position_ego
            )
            distance_lanelets.append((id_lanelet, p_curvilinear_long, distance, ccosy))

        if not distance_lanelets:
            raise BaseException(
                f"ego vehicle with position {position_ego}"
                f"has no closest lanelet in lanelet ids: {self.route.list_ids_lanelets}"
            )

        distance_lanelets.sort(
            key=lambda tup: abs(tup[2])
        )  # sorts in place by distance

        # LOGGER.error(distance_lanelets)

        idx_closest_lanelet = self.route.list_ids_lanelets.index(
            distance_lanelets[0][0]
        )
        for count, rel_id in enumerate(lanelets_id_rel):
            # rel_id is 0 for closest, +1 for successor of closest, +-1 for predecessor of closest
            idx_observe = min(
                max(rel_id + idx_closest_lanelet, 0),
                len(self.route.list_ids_lanelets) - 1,
            )  # padding to first / last if this position is out of bound

            lanelet_id_observe = self.route.list_ids_lanelets[idx_observe]

            id_lane, p_curvi_long, distance_lat, ccosy = next(
                (id_lane, p_curvi_long, distance_lat, ccosy)
                for (id_lane, p_curvi_long, distance_lat, ccosy) in distance_lanelets
                if id_lane == lanelet_id_observe
            )

            # make obervation, where the predecessor lanelet is.

            vectors_to_ref, orientations_ref = self.get_points_on_ccosy(
                ccosy,
                p_curvi_long,
                distances_per_lanelet,
                orientation_vehicle,
                position_ego,
            )
            # LOGGER.debug(
            #     f"observe points {vectors_to_ref} and orients {orientations_ref} for lanelet {lanelet_id_observe}, "
            #     f"which is at position {rel_id} to closest lanelet {distance_lanelets[0][0]} "
            #     f"at ego position {position_ego}"
            # )
            computed_waypoints[count, :, :] = vectors_to_ref
            computed_orientations[count, :] = orientations_ref

        return (computed_waypoints, computed_orientations)

    @staticmethod
    def sorted_lanelet_ids(
        lanelet_ids: List[int],
        orientation: float,
        position: np.ndarray,
        scenario: Scenario,
    ) -> List[int]:
        """
        return the lanelets sorted by relative orientation to the position and orientation given
        """
        # version 2020
        if len(lanelet_ids) <= 1:
            return lanelet_ids
        else:
            lanelet_id_list = np.array(lanelet_ids)

            def get_lanelet_relative_orientation(lanelet_id):
                lanelet = scenario.lanelet_network.find_lanelet_by_id(lanelet_id)
                lanelet_orientation = Navigator.lanelet_orientation_at_position(
                    lanelet, position
                )
                return np.abs(
                    Navigator.relative_orientation(lanelet_orientation, orientation)
                )

            orientation_differences = np.array(
                list(map(get_lanelet_relative_orientation, lanelet_id_list))
            )
            sorted_indices = np.argsort(orientation_differences)
            return list(lanelet_id_list[sorted_indices])

    @staticmethod
    def lanelet_orientation_at_position(lanelet: Lanelet, position: np.ndarray):
        """
        Approximates the lanelet orientation with the two closest point to the given state
        # TODO optimize more for speed

        :param lanelet: Lanelet on which the orientation at the given state should be calculated
        :param position: Position where the lanelet's orientation should be calculated
        :return: An orientation in interval [-pi,pi]
        """
        # version 2020

        center_vertices = lanelet.center_vertices

        position_diff = []
        for idx in range(len(center_vertices) - 1):
            vertex1 = center_vertices[idx]
            position_diff.append(np.linalg.norm(position - vertex1))

        closest_vertex_index = position_diff.index(min(position_diff))

        vertex1 = center_vertices[closest_vertex_index, :]
        vertex2 = center_vertices[closest_vertex_index + 1, :]
        direction_vector = vertex2 - vertex1
        return np.arctan2(direction_vector[1], direction_vector[0])

    @staticmethod
    def relative_orientation(from_angle1_in_rad, to_angle2_in_rad):
        # version 2020
        phi = (to_angle2_in_rad - from_angle1_in_rad) % (2 * np.pi)
        if phi > np.pi:
            phi -= 2 * np.pi

        return phi



"""Module containing the Termination class"""
from commonroad_rl.gym_commonroad.action import Action


class Termination:
    """Class for detecting if the scenario should be terminated"""

    def __init__(self, config: dict):
        """
        :param config: Configuration of the environment
        """

        self.termination_configs = config["termination_configs"]
        self.num_friction_violation = 0

    def reset(self, observation: dict, ego_action: Action):
        """
        For resetting the memory the termination class has for the current scenario

        :param observation: Current observation of the environment
        :param ego_action: Current ego_action of the environment
        """
        self.num_friction_violation = 0

    def is_terminated(self, observation: dict, ego_action: Action) -> (bool, str, dict):
        """
        Detect if the scenario should be terminated

        :param observation: Current observation of the environment
        :param ego_action: Current ego_action of the environment
        :return: Tuple of (terminated: bool, reason: str, termination_info: dict)
        """

        done = False
        termination_info = {
            "is_goal_reached": 0,
            "is_collision": 0,
            "is_off_road": 0,
            "is_time_out": 0,
            "is_friction_violation": 0,
            "num_friction_violation": self.num_friction_violation
        }
        termination_reason = None

        if observation["is_off_road"][0]:  # Ego vehicle is off-road
            termination_info["is_off_road"] = 1
            if self.termination_configs["terminate_on_off_road"]:
                done = True
                termination_reason = "is_off_road"

        elif observation["is_collision"][0]:  # Collision with others
            termination_info["is_collision"] = 1
            if self.termination_configs["terminate_on_collision"]:
                done = True
                termination_reason = "is_collision"

        elif observation["is_goal_reached"][0]:  # Goal region is reached
            termination_info["is_goal_reached"] = 1
            if self.termination_configs["terminate_on_goal_reached"]:
                done = True
                termination_reason = "is_goal_reached"

        elif observation["is_time_out"][0]:  # Max simulation time step is reached
            termination_info["is_time_out"] = 1
            if self.termination_configs["terminate_on_time_out"]:
                done = True
                termination_reason = "is_time_out"

        elif observation["is_friction_violation"][0]:  # Friction limitation is violated
            self.num_friction_violation += 1
            termination_info["is_friction_violation"] = 1
            termination_info["num_friction_violation"] = self.num_friction_violation
            if self.termination_configs["terminate_on_friction_violation"]:
                done = True
                termination_reason = "is_friction_violation"

        return done, termination_reason, termination_info



"""
Module for plotting learning curves
"""
import os
os.environ["KMP_WARNINGS"] = "off"
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"
import logging

logging.getLogger("tensorflow").disabled = True

import matplotlib

matplotlib.use("pdf")
import matplotlib.pyplot as plt

import numpy as np
import math
import argparse
from commonroad_rl.utils_run.plot_util import smooth
from commonroad_rl.utils_run.plot_util import plot_results as plot_results_baselines
from commonroad_rl.utils_run.plot_util import load_results as load_results_baselines

try:
    from gym_monitor.util import *
except:
    pass

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
LOGGER.addHandler(handler)

LATEX = False
LABELPAD = 15
FIGSIZE = (14, 12)

if LATEX:
    # use Latex font
    FONTSIZE = 28
    plt.rcParams['text.latex.preamble'] = [r"\usepackage{lmodern}"]
    pgf_with_latex = {  # setup matplotlib to use latex for output
        "pgf.texsystem": "pdflatex",  # change this if using xetex or lautex
        "text.usetex": True,  # use LaTeX to write all text
        "font.family": 'lmodern',
        # blank entries should cause plots
        "font.sans-serif": [],  # ['Avant Garde'],              # to inherit fonts from the document
        # 'text.latex.unicode': True,
        "font.monospace": [],
        "axes.labelsize": FONTSIZE,  # LaTeX default is 10pt font.
        "font.size": FONTSIZE - 10,
        "legend.fontsize": FONTSIZE,  # Make the legend/label fonts
        "xtick.labelsize": FONTSIZE,  # a little smaller
        "ytick.labelsize": FONTSIZE,
        "pgf.preamble": [
            r"\usepackage[utf8x]{inputenc}",  # use utf8 fonts
            r"\usepackage[T1]{fontenc}",  # plots will be generated
            r"\usepackage[detect-all,locale=DE]{siunitx}",
        ]  # using this preamble
    }
    matplotlib.rcParams.update(pgf_with_latex)


def argsparser():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-f", "--log-folder", help="Log folder", type=str, default="log")
    parser.add_argument("--model_path", "-model", type=str, nargs="+", default=(),
                        help="(tuple) Relative path of the to be plotted model from the log folder")
    parser.add_argument("--no_render", "-nr", action="store_true", help="Whether to render images")
    parser.add_argument("-t", "--title", help="Figure title", type=str, default="result")
    # TODO: integrate sliding window size
    parser.add_argument("--smooth", action="store_true",
                        help="Smooth learning curves (average around a sliding window)")
    parser.add_argument("--format", "-file", help="Image format to save", type=str, default="png")
    parser.add_argument("-ic", "--invalid_collision", action="store_true",
                        help="Plot invalid collisions only graph " +
                             "- requires logging of info-keyword valid_collision")
    parser.add_argument("-tc", "--total_collisions", action="store_true",
                        help="Plot total (valid + invalid) collisions graph " +
                             "- requires logging of info-keyword valid_collision")

    return parser.parse_args()


def ts2reward(results):
    """
    Decompose a timesteps variable to x ans ys

    :param timesteps: (Pandas DataFrame) the input data
    :param xaxis: (str) the axis for the x and y output
        (can be X_TIMESTEPS='timesteps', X_EPISODES='episodes' or X_WALLTIME='walltime_hrs')
    :return: (np.ndarray, np.ndarray) the x and y output
    """

    x_var = np.cumsum(results.monitor.l)
    y_var = results.monitor.r.values

    return x_var, y_var


def ts2goal(results):
    """
    Decompose a timesteps variable to x ans ys

    :param timesteps: (Pandas DataFrame) the input data
    :param xaxis: (str) the axis for the x and y output
        (can be X_TIMESTEPS='timesteps', X_EPISODES='episodes' or X_WALLTIME='walltime_hrs')
    :return: (np.ndarray, np.ndarray) the x and y output
    """

    x_var = np.cumsum(results.monitor.l.values)
    y_var = results.monitor.is_goal_reached.values

    return x_var, y_var


def ts2collision(results):
    """
    Decompose a timesteps variable to x ans ys

    :param timesteps: (Pandas DataFrame) the input data
    :param xaxis: (str) the axis for the x and y output
        (can be X_TIMESTEPS='timesteps', X_EPISODES='episodes' or X_WALLTIME='walltime_hrs')
    :return: (np.ndarray, np.ndarray) the x and y output
    """

    x_var = np.cumsum(results.monitor.l)
    if hasattr(results.monitor, "valid_collision"):
        y_var = results.monitor.valid_collision
    else:
        y_var = results.monitor.is_collision
    """
    else:
        tmp = results.monitor.termination_reason != "invalid_collision"
        y_var = np.logical_and(np.logical_and(
            results.monitor.is_collision.values,
            np.logical_not(results.monitor.is_infeasible.values)),
            tmp.values
        )
    """

    return x_var, y_var


def ts2off_road(results):
    """
    Decompose a timesteps variable to x ans ys

    :param timesteps: (Pandas DataFrame) the input data
    :param xaxis: (str) the axis for the x and y output
        (can be X_TIMESTEPS='timesteps', X_EPISODES='episodes' or X_WALLTIME='walltime_hrs')
    :return: (np.ndarray, np.ndarray) the x and y output
    """

    x_var = np.cumsum(results.monitor.l)
    if hasattr(results.monitor, "valid_off_road"):
        y_var = results.monitor.valid_off_road.values
    else:
        y_var = results.monitor.is_off_road.values
    """
    else:
        y_var = np.logical_and(results.monitor.is_off_road.values, np.logical_not(results.monitor.is_infeasible.values))
    """

    return x_var, y_var


def ts2max_time(results):
    """
    Decompose a timesteps variable to x ans ys

    :param timesteps: (Pandas DataFrame) the input data
    :param xaxis: (str) the axis for the x and y output
        (can be X_TIMESTEPS='timesteps', X_EPISODES='episodes' or X_WALLTIME='walltime_hrs')
    :return: (np.ndarray, np.ndarray) the x and y output
    """
    x_var = np.cumsum(results.monitor.l.values)
    y_var = results.monitor.is_time_out.values

    return x_var, y_var


def ts2goal_time(results):
    """
    Decompose a timesteps variable to x ans ys

    :param timesteps: (Pandas DataFrame) the input data
    :param xaxis: (str) the axis for the x and y output
        (can be X_TIMESTEPS='timesteps', X_EPISODES='episodes' or X_WALLTIME='walltime_hrs')
    :return: (np.ndarray, np.ndarray) the x and y output
    """

    x_var = np.cumsum(results.monitor.l)
    y_var = [0]
    for i, is_goal_reached in enumerate(results.monitor.is_goal_reached.values):
        if is_goal_reached:
            y_var.append(results.monitor.current_episode_time_step.values[i])
        else:
            # y_var.append(results.monitor.max_episode_time_steps[i])
            y_var.append(y_var[-1])

    return x_var, np.array(y_var[1:])


def ts2infeasible(results):
    """
    Decompose a timesteps variable to x ans ys

    :param timesteps: (Pandas DataFrame) the input data
    :param xaxis: (str) the axis for the x and y output
        (can be X_TIMESTEPS='timesteps', X_EPISODES='episodes' or X_WALLTIME='walltime_hrs')
    :return: (np.ndarray, np.ndarray) the x and y output
    """

    x_var = np.cumsum(results.monitor.l)
    y_var = results.monitor.is_infeasible.values

    return x_var, y_var


# TODO: plot individual infeasible reasons
def ts2LongitudinalInfeasible(results):
    x_var = np.cumsum(results.monitor.l)
    y_var = results.monitor.infeasible_reason.values == "LongitudinalInfeasible"

    return x_var, y_var


def ts2LateralInfeasible(results):
    x_var = np.cumsum(results.monitor.l)
    y_var = results.monitor.infeasible_reason.values == "LateralInfeasible"

    return x_var, y_var


def ts2NoLongDrivingCorridors(results):
    x_var = np.cumsum(results.monitor.l)
    y_var = results.monitor.infeasible_reason.values == "NoLongDrivingCorridors"

    return x_var, y_var


def ts2num_safe_actions(results):
    x_var = np.cumsum(results.monitor.l)
    y_var = results.monitor.num_safe_actions.values #/results.monitor.l

    return x_var, y_var

def ts2unknown_braking(results):
    x_var = np.cumsum(results.monitor.l)
    y_var = results.monitor.infeasible_reason.values == "unknown_braking"

    return x_var, y_var


def ts2friction_violation(results):
    """
    Decompose a timesteps variable to x ans ys

    :param timesteps: (Pandas DataFrame) the input data
    :param xaxis: (str) the axis for the x and y output
        (can be X_TIMESTEPS='timesteps', X_EPISODES='episodes' or X_WALLTIME='walltime_hrs')
    :return: (np.ndarray, np.ndarray) the x and y output
    """

    x_var = np.cumsum(results.monitor.l.values)
    y_var = results.monitor.is_friction_violation.values

    return x_var, y_var


def ts2v_ego(results):
    # TODO: plot percentile
    # x_var = np.cumsum(results.monitor.l.values) * 1e-3
    # y_var = np.abs(results.monitor.v_ego_mean.values)
    n_percentile = np.zeros((11))
    v_ego_mean = np.array(results.monitor.v_ego_mean.values)
    for p in range(11):
        n_percentile[p] = np.percentile(v_ego_mean, p * 10)

    return np.array(range(11)), n_percentile


def ts2u1(results):
    """
    Decompose a timesteps variable to x ans ys

    :param timesteps: (Pandas DataFrame) the input data
    :param xaxis: (str) the axis for the x and y output
        (can be X_TIMESTEPS='timesteps', X_EPISODES='episodes' or X_WALLTIME='walltime_hrs')
    :return: (np.ndarray, np.ndarray) the x and y output
    """

    x_var = np.cumsum(results.monitor.l.values)
    y_var = np.abs(results.monitor.u_cbf_1_sum.values) / results.monitor.l.values
    return x_var, y_var


def ts2u2(results):
    """
    Decompose a timesteps variable to x ans ys

    :param timesteps: (Pandas DataFrame) the input data
    :param xaxis: (str) the axis for the x and y output
        (can be X_TIMESTEPS='timesteps', X_EPISODES='episodes' or X_WALLTIME='walltime_hrs')
    :return: (np.ndarray, np.ndarray) the x and y output
    """

    x_var = np.cumsum(results.monitor.l.values)
    y_var = np.abs(results.monitor.u_cbf_1_sum.values) / results.monitor.l.values
    return x_var, y_var


PLOT_DICT = {
    "Total Reward": ts2reward,
    "Goal-Reaching Rate": ts2goal,
    "Valid Collision Rate": ts2collision,
    "Off-Road Rate": ts2off_road,
    "Time-Out Rate": ts2max_time,
    "Infeasible Rate": ts2infeasible,
    "LongitudinalInfeasible Rate": ts2LongitudinalInfeasible,
    "LateralInfeasible Rate": ts2LateralInfeasible,
    "NoLongDrivingCorridors Rate": ts2NoLongDrivingCorridors,
    "Total Num Safe Actions": ts2num_safe_actions,
    # "Goal Reaching Time": ts2goal_time,
    # "Mean ego velocity": ts2v_ego,
    # "Total Robustness reward": ts2monitor_reward,
    # "Total Sparse reward": ts2gym_reward,
    # "Min Robustness": ts2min_robustness,
    # "Max Robustness": ts2max_robustness,
    # "Avg. Step Robustness reward": ts2monitor_reward_step,
    # "Avg. Step Sparse reward": ts2gym_reward_step,
    # "True traffic rule violation": ts2rule_violation,
    # "Valid traffic rule violation": ts2valid_rule_violation,
    # "Active step robustness reward": ts2active_step_robustness,
    # "Active total robustness reward": ts2active_total_robustness,
    # "Step robustness reward vs num violation": violation2step_robustness_tmp,
    # "$|u_1 - u_\mathrm{RL1}|$   [rad/$\mathrm{s}^2$]": ts2u1,
    # "$|u_2 - u_\mathrm{RL2}|$   [m/$\mathrm{s}^2$]": ts2u2
    # "Friction violation": ts2friction_violation,
}


def group_fn(results):
    return os.path.basename(results.dirname)


def main():
    args = argsparser()

    if args.invalid_collision:
        PLOT_DICT["Invalid Collision Rate"] = ts2invalidcollision
    if args.total_collisions:
        PLOT_DICT["Total Collision Rate"] = ts2totalcollision

    log_dir = args.log_folder
    model_paths = tuple(args.model_path)

    num_of_columns = 2
    num_of_rows = math.ceil(len(PLOT_DICT) / num_of_columns)

    for idx, model in enumerate(model_paths):
        fig, axarr = plt.subplots(num_of_rows, num_of_columns, sharex=False, squeeze=True, figsize=FIGSIZE, dpi=100)
        results = load_results_baselines(os.path.join(log_dir, model))

        for i, (k, xy_fn) in enumerate(PLOT_DICT.items()):
            legend = i == 0
            plot_line = False  # i >1
            set_y_lim = "Rate" in k
            try:
                fig, axarr = plot_results_baselines(
                    results, fig, axarr,
                    nrows=num_of_rows, ncols=num_of_columns, xy_fn=xy_fn,
                    idx_row=i // num_of_columns, idx_col=i % num_of_columns,
                    average_group=False, resample=args.smooth,
                    group_fn=group_fn,
                    xlabel="Training Steps * 1e6", ylabel=k,
                    # xlabel="\\textbf{Training Steps * 1000}",
                    # ylabel="\\textbf{" + k + "}",
                    labelpad=LABELPAD,
                    legend_outside=True,
                    legend=legend, plot_line=plot_line, set_y_lim=set_y_lim)
            except AttributeError:
                continue
        plt.tight_layout()
        fig.savefig(os.path.join(log_dir, model, f"{model}.{args.format}"), format=args.format, bbox_inches='tight')
        LOGGER.info(f"Saved {model}.{args.format} to {log_dir}/{model}")  # , figure, os.path.join(log_dir, model))


if __name__ == "__main__":
    main()



""" Provides a function to classify a trajectory based on its curvature profile
"""

import math
import os
from enum import Enum

import numpy as np
from commonroad.scenario.trajectory import Trajectory


class TrajectoryType(Enum):
    LEFT = 1
    RIGHT = 2
    STRAIGHT = 3
    BOTH = 4


def classify_trajectory(
    t: Trajectory, min_velocity=1.0, turn_threshold=0.03
):
    """
    Get TrajectoryType of the given trajectory
    :param t: trajectory to classify
    :param min_velocity: curvatures for states with velocity < min_velocity are clamped to 0
    :param turn_threshold: minimum curvature at any point for trajectory to be classified as a turn
    :return: a goal shape of area larger than the ego vehicle,
    """

    v = [s.velocity for s in t.state_list]
    c = np.array(_calc_curvature(t))
    c[np.abs(v) < min_velocity] = 0.0
    c = _smooth(c, 2, 13)

    traj_class = _classify_curvature(c, turn_threshold)

    return traj_class, c


def _classify_curvature(c, turn_threshold):
    """Get TrajectoryType based off curvatures"""

    min_c = min(c)
    max_c = max(c)

    is_right_turn = min_c <= -turn_threshold
    is_left_turn = max_c >= turn_threshold

    if is_left_turn and not is_right_turn:
        return TrajectoryType.LEFT
    elif is_right_turn and not is_left_turn:
        return TrajectoryType.RIGHT
    elif not is_left_turn and not is_right_turn:
        return TrajectoryType.STRAIGHT
    else:
        return TrajectoryType.BOTH


def _smooth(x, iterations=1, window=2):

    if iterations <= 0:
        return x

    x_new = x[:]
    for i in range(len(x)):
        lb = max(0, i - window)
        ub = min(len(x) - 1, i + window)
        x_new[i] = sum(x[lb:ub]) / (ub - lb)

    return _smooth(x_new, iterations - 1, window)


def _calc_derivative(x, t):
    if len(x) <= 2:
        return len(x) * [x[0]]
    dx = []
    for i in range(1, len(x) - 1):
        dx.append((x[i + 1] - x[i - 1]) / (t[i + 1] - t[i - 1]))
    # pad start and end using linear extrapolation to get same length as before
    if len(dx) > 1:
        dx.append(dx[-1] + (dx[-1] - dx[-2]))
        dx.insert(0, dx[0] + (dx[0] - dx[1]))
    elif len(dx) == 1:
        dx.append(dx[-1])
        dx.append(dx[-1])
    return dx


def _calc_curvature(traj: Trajectory):

    t = list(map(lambda s: s.time_step, traj.state_list))
    x = list(map(lambda s: s.position[0], traj.state_list))
    y = list(map(lambda s: s.position[1], traj.state_list))

    x1 = _calc_derivative(x, t)
    y1 = _calc_derivative(y, t)
    x2 = _calc_derivative(x1, t)
    y2 = _calc_derivative(y1, t)

    c = map(
        lambda a: (a[0] * a[3] - a[2] * a[1]) / ((a[0] ** 2 + a[2] ** 2) ** 1.5) if (a[0]**2 + a[2] ** 2) ** 1.5 != 0
        else (a[0] * a[3] - a[2] * a[1]) / 1E-7,
        zip(x1, x2, y1, y2),
    )
    c = list(c)

    c = _smooth(c, 2, 5)

    return c


# """
# Module containing the observation base class
# """
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Union, Dict

import numpy as np
from commonroad.scenario.obstacle import State


class Observation(ABC):
    """
    Abstract class to define an observation
    """

    @abstractmethod
    def observe(self, ego_state: State, **kwargs) -> Union[np.array, Dict]:
        """ Create observation for given state in an environment.

            :param ego_state: state from which to observe the environment
            :return: ndarray of observation if flatten == True, observation dict otherwise
        """
        pass

    @abstractmethod
    def draw(self, render_configs: Dict, **kwargs):
        """ Method to draw the observation """
        pass

    @abstractmethod
    def build_observation_space(self) -> OrderedDict:
        """ Method to build the observation space """
        pass



"""
Loads a set of pickled scenarios and measures loading time and memory consumption
"""
import os
import pickle
import glob
from time import time
from argparse import ArgumentParser
from memory_profiler import profile


def load_scenarios(path):
    storage = []

    meta_scenario_reset_dict_path = os.path.join(
        path, "meta_scenario", "meta_scenario_reset_dict.pickle"
    )
    with open(meta_scenario_reset_dict_path, "rb") as f:
        storage.append(pickle.load(f))

    problem_meta_scenario_dict_path = os.path.join(
        path, "meta_scenario", "problem_meta_scenario_dict.pickle"
    )
    with open(problem_meta_scenario_dict_path, "rb") as f:
        storage.append(pickle.load(f))

    print(f"Loading {path}...")
    fns = glob.glob(os.path.join(path, "problem_train", "*.pickle"))
    for fn in fns:
        with open(fn, "rb") as f:
            storage.append(pickle.load(f))

    return len(storage) - 2


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("path", type=str, help="Path to pickle training data")
    parser.add_argument(
        "-m",
        "--memory",
        action="store_true",
        help="Do memory profiling of the pickle loading",
    )
    args = parser.parse_args()

    _begin = time()
    if args.memory:
        load_scenarios = profile(load_scenarios)
    num = load_scenarios(args.path)
    _end = time()
    print(f"Took {_end-_begin}s to load {num} scenarios")



"""
Module for playing trained model using Stable baselines
"""
import argparse
import csv
import glob
import logging
import os
os.environ["KMP_WARNINGS"] = "off"
os.environ["KMP_AFFINITY"] = "none"
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"
logging.getLogger("tensorflow").disabled = True
import pickle
from typing import Union
import numpy as np
import yaml
from gym import Env
from stable_baselines.common import set_global_seeds

from commonroad_rl.gym_commonroad.commonroad_env import CommonroadEnv
from commonroad_rl.gym_commonroad.constants import PATH_PARAMS
from commonroad_rl.train_model import LoggingMode
from commonroad_rl.utils_run.utils import get_wrapper_class, make_env, load_model_and_vecnormalize
from commonroad_rl.utils_run.vec_env import CommonRoadVecEnv

try:
    from mpi4py import MPI
except ImportError:
    print("ImportFailure MPI")
    MPI = None

LOGGER = logging.getLogger(__name__)


def get_parser():
    parser = argparse.ArgumentParser(description="Evaluates PPO2 trained model with specified test scenarios",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--env_id", type=str, default="commonroad-v1", help="environment ID")
    parser.add_argument("--algo", type=str, default="ppo2")
    parser.add_argument("--scenario_path", "-i", type=str, help="Path to pickled test scenarios",
                        default=PATH_PARAMS["test_reset_config"])
    parser.add_argument("--test_folder", type=str, help="Folder for problem pickles", default="problem_test")
    parser.add_argument("--model_path", "-model", type=str, help="Path to trained model",
                        default=PATH_PARAMS["log"] + "/ppo2/commonroad-v0_3")
    parser.add_argument("--evaluation_path", "-eval_path", type=str, help="Folder to store evaluation data",
                        default="evaluation")
    parser.add_argument("--viz_path", "-viz", type=str, default="")
    parser.add_argument("--num_scenarios", "-n", default=-1, type=int, help="Maximum number of scenarios to draw")
    parser.add_argument("--multiprocessing", "-mpi", action="store_true")
    parser.add_argument("--combine_frames", "-1", action="store_true",
                        help="Combine rendered environments into one picture")
    parser.add_argument("--skip_timesteps", "-st", type=int, default=1,
                        help="Only render every nth frame (including first and last)")
    parser.add_argument("--no_render", "-nr", action="store_true", help="Whether store render images")
    parser.add_argument("--no_csv", action="store_true", help="Not store statistic results to speed up MPI processes")
    parser.add_argument("--hyperparam_filename", "-hyperparam_f", type=str, default="model_hyperparameters.yml")
    parser.add_argument("--config_filename", "-config_f", type=str, default="environment_configurations.yml")
    parser.add_argument("--log_action_curve", action="store_true", help="Store action curve plot for analysis")
    parser.add_argument("--log_step_info", action="store_true", help="Store all info dict in step function")
    parser.add_argument("--logging_mode", default=LoggingMode.INFO, type=LoggingMode, choices=list(LoggingMode))

    return parser


def create_environments(env_id: str, model_path: str, viz_path: str,
                        hyperparam_filename: str = "model_hyperparameters.yml",
                        config_filename: str = "environment_configurations.yml",
                        logging_mode: LoggingMode = LoggingMode.INFO.value, test_path=None, meta_path=None,
                        play: bool = True, test_env: bool = True, **kwargs) -> CommonRoadVecEnv:
    """
    Create CommonRoad vectorized environment environment

    :param config_filename:
    :param env_id: Environment gym id
    :param test_path: Path to the test files
    :param meta_path: Path to the meta-scenarios
    :param model_path: Path to the trained model
    :param viz_path: Output path for rendered images
    :param hyperparam_filename: The filename of the hyperparameters
    :param env_kwargs: Keyword arguments to be passed to the environment
    """
    # Get environment keyword arguments including observation and reward configurations
    config_fn = os.path.join(model_path, config_filename)
    env_kwargs = {"logging_mode": logging_mode,
                  "visualization_path": viz_path,
                  "play": play,
                  "test_env": test_env,
                  "config_file": config_fn}

    env_kwargs.update(kwargs)
    if meta_path is not None:
        env_kwargs.update({"meta_scenario_path": meta_path})
    if test_path is not None:
        env_kwargs.update({"test_reset_config_path": test_path})

    # Load model hyperparameters:
    hyperparam_fn = os.path.join(model_path, hyperparam_filename)
    with open(hyperparam_fn, "r") as f:
        hyperparams = yaml.load(f, Loader=yaml.Loader)

    env_wrapper = get_wrapper_class(hyperparams)

    # Create environment
    # note that CommonRoadVecEnv is inherited from DummyVecEnv
    env = CommonRoadVecEnv([make_env(env_id, 0, wrapper_class=env_wrapper, env_kwargs=env_kwargs, subproc=False)])

    def on_reset_callback(env: Union[Env, CommonroadEnv], elapsed_time: float):
        # reset callback called before resetting the env
        if env.observation_dict["is_goal_reached"][-1]:
            LOGGER.debug("Goal reached")
        else:
            LOGGER.debug("Goal not reached")
        # env.render()

    env.set_on_reset(on_reset_callback)

    return env, hyperparams["normalize"]


def main():
    args = get_parser().parse_args()

    if MPI is None:
        args.multiprocessing = False

    LOGGER.setLevel(args.logging_mode.value)
    handler = logging.StreamHandler()
    handler.setLevel(args.logging_mode.value)
    LOGGER.addHandler(handler)

    meta_path = os.path.join(args.scenario_path, "meta_scenario")

    # mpi for parallel processing
    rank = 0
    size = 1
    comm = None
    if args.multiprocessing:
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        size = comm.Get_size()
        test_path = os.path.join(args.scenario_path, args.test_folder, str(rank))

    else:
        test_path = os.path.join(args.scenario_path, args.test_folder)

    # create evaluation folder in model_path
    evaluation_path = os.path.join(args.model_path, args.evaluation_path)
    os.makedirs(evaluation_path, exist_ok=True)
    if args.viz_path == "":
        args.viz_path = os.path.join(evaluation_path, "img")

    env, normalize = create_environments(
        args.env_id, args.model_path, args.viz_path, args.hyperparam_filename, args.config_filename,
        args.logging_mode.value, test_path=test_path, meta_path=meta_path)

    LOGGER.info(f"Testing a maximum of {args.num_scenarios} scenarios")

    model, env = load_model_and_vecnormalize(args.model_path, args.algo, normalize, env)

    set_global_seeds(1)
    num_valid_collisions, num_collisions, num_valid_off_road, num_off_road, num_goal_reaching, \
        num_timeout, total_scenarios = 0, 0, 0, 0, 0, 0, 0

    if args.log_action_curve:
        accelerations = {}
        jerks = {}

    # In case there a no scenarios at all
    try:
        obs = env.reset()
    except IndexError:
        args.num_scenarios = 0

    count = 0
    info_dict = {}

    while count != args.num_scenarios:
        done = False
        if not args.no_render:
            env.render()
        benchmark_id = env.venv.envs[0].benchmark_id
        LOGGER.debug(benchmark_id)
        if args.log_action_curve:
            accelerations[benchmark_id] = []
            jerks[benchmark_id] = []
        while not done:
            action, _states = model.predict(obs, deterministic=True)
            obs, reward, done, info = env.step(action)
            if not args.no_render:
                env.render()
            if args.log_step_info:
                for kwargs in info[0]:
                    if kwargs not in info_dict:
                        info_dict[kwargs] = [info[0][kwargs]]
                    else:
                        info_dict[kwargs].append(info[0][kwargs])

            if done:
               LOGGER.info(f'scenario={info[0]["scenario_name"]}, termination_reason={info[0]["termination_reason"]}')

            if args.log_action_curve:
                dt = env.venv.envs[0].scenario.dt
                ego_vehicle = env.venv.envs[0].ego_action.vehicle
                jerk_x = (ego_vehicle.state.acceleration - ego_vehicle.previous_state.acceleration) / dt
                jerk_y = (ego_vehicle.state.acceleration_y - ego_vehicle.previous_state.acceleration_y) / dt
                accelerations[benchmark_id].append(
                    [ego_vehicle.state.acceleration, ego_vehicle.state.acceleration_y])
                jerks[benchmark_id].append([jerk_x, jerk_y])

        # log collision rate, off-road rate, and goal-reaching rate
        info = info[0]

        total_scenarios += 1
        num_valid_collisions += info.get("valid_collision", info["is_collision"])
        num_collisions += info["is_collision"]
        num_timeout += info["is_time_out"]
        num_valid_off_road += info.get("valid_off_road", info["is_off_road"])
        num_off_road += info["is_off_road"]
        num_goal_reaching += info["is_goal_reached"]

        out_of_scenarios = info["out_of_scenarios"]
        if not args.no_csv:
            with open(os.path.join(evaluation_path, f"{rank}_results.csv"), "a") as f:
                csv_writer = csv.writer(f)
                csv_writer.writerow((info["scenario_name"], info["current_episode_time_step"], info["termination_reason"]))

        if out_of_scenarios:
            break
        count += 1

    if args.log_action_curve:
        for key in accelerations.keys():
            accelerations[key] = np.array(accelerations[key])
            jerks[key] = np.array(jerks[key])
        with open(f"{args.model_path.split('/')[-1]}_actions.pkl", 'wb') as f:
            pickle.dump({"accelerations": accelerations, "jerks": jerks}, f)

    if args.log_step_info:
        with open(f"{os.path.abspath(args.model_path)}_step_info_{rank}.pkl", 'wb') as f:
            pickle.dump(info_dict, f)
    if args.multiprocessing:
        data = (total_scenarios, num_valid_collisions, num_collisions, num_valid_off_road, num_off_road, \
                num_timeout, num_goal_reaching)
        data = comm.gather(data)
        if rank == 0:
            g_num_scenarios, g_num_valid_collisions, g_num_collisions, g_num_valid_off_road, \
                g_num_off_road, g_num_timeout, g_num_goal_reaching = zip(*data)
            total_scenarios = sum(g_num_scenarios)
            num_valid_collisions = sum(g_num_valid_collisions)
            num_collisions = sum(g_num_collisions)
            num_valid_off_road = sum(g_num_valid_off_road)
            num_off_road = sum(g_num_off_road)
            num_timeout = sum(g_num_timeout)
            num_goal_reaching = sum(g_num_goal_reaching)
        else:
            return
    if not args.no_csv:
        # save evaluation results
        with open(os.path.join(evaluation_path, "results.csv"), 'w') as fd_result:
            fd_result.write("benchmark_id, time_steps, termination_reason\n")
            for i in range(size):
                path = os.path.join(evaluation_path, f"{i}_results.csv")
                with open(path, 'r') as f:
                    fd_result.write(f.read())
                os.remove(path)

        with open(os.path.join(evaluation_path, "overview.yml"), "w") as f:
            yaml.dump({
                "total_scenarios": total_scenarios,
                "num_collisions": num_collisions,
                "num_valid_collisions": num_valid_collisions,
                "num_valid_off_road": num_valid_off_road,
                "num_timeout": num_timeout,
                "num_off_road": num_off_road,
                "num_goal_reached": num_goal_reaching,
                "percentage_goal_reached": 100.0 * num_goal_reaching / total_scenarios,
                "percentage_off_road": 100.0 * num_off_road / total_scenarios,
                "percentage_collisions": 100.0 * num_collisions / total_scenarios,
                "percentage_valid_off_road": 100.0 * num_valid_off_road / total_scenarios,
                "percentage_valid_collisions": 100.0 * num_valid_collisions / total_scenarios,
                "percentage_timeout": 100.0 * num_timeout / total_scenarios
            }, f)

    if not args.no_csv:
        # TODO: fix for no_render mode
        # Reorganize the rendered images according to result of the scenario
        # Flatten directory
        if not args.no_render:
            img_path = args.viz_path
            for d in os.listdir(img_path):
                dir_path = os.path.join(img_path, d)
                if not os.path.isdir(dir_path):
                    continue

                for f in os.listdir(dir_path):
                    os.rename(os.path.join(dir_path, f), os.path.join(img_path, f))

                os.rmdir(dir_path)

        # Split into different termination reasons
            with open(os.path.join(evaluation_path, "results.csv"), "r") as f:
                os.mkdir(os.path.join(img_path, "time_out"))
                os.mkdir(os.path.join(img_path, "off_road"))
                os.mkdir(os.path.join(img_path, "collision"))
                os.mkdir(os.path.join(img_path, "collision", "valid_collision"))
                os.mkdir(os.path.join(img_path, "collision", "collision_caused_by_other_vehicle"))
                os.mkdir(os.path.join(img_path, "goal_reached"))
                os.mkdir(os.path.join(img_path, "other"))

                reader = csv.reader(f)
                reader.__next__()
                for [scenario_id, _, t_reason] in reader:
                    if args.no_render:
                        if t_reason == "valid_collision" or t_reason == "collision_caused_by_other_vehicle":
                            dest_path = os.path.join(img_path,"collision",t_reason)
                        else:
                            dest_path = os.path.join(img_path, t_reason)
                    else:
                        if t_reason == "valid_collision" or t_reason == "collision_caused_by_other_vehicle":
                            os.mkdir(os.path.join(img_path, "collision", t_reason, scenario_id))
                            dest_path = os.path.join(img_path, "collision", t_reason, scenario_id)
                        else:
                            os.mkdir(os.path.join(img_path, t_reason, scenario_id))
                            dest_path = os.path.join(img_path, t_reason, scenario_id)
                    for file_path in glob.glob(os.path.join(img_path, scenario_id + '*')):
                        os.rename(file_path, os.path.join(dest_path, os.path.basename(file_path)))


if __name__ == "__main__":
    main()


"""Class for hybrid reward"""
import logging
from math import exp

import numpy as np
from commonroad.common.solution import VehicleModel

from commonroad_rl.gym_commonroad.action import Action
from commonroad_rl.gym_commonroad.reward.reward import Reward

LOGGER = logging.getLogger(__name__)


class HybridReward(Reward):
    """Class for hybrid reward"""

    def __init__(self, configs: dict):
        self.goal_configs = configs["goal_configs"]
        self.reward_configs = configs["reward_configs"]["hybrid_reward"]
        #        self.traffic_sign_configs = configs["traffic_sign_configs"]
        self.lanelet_configs = configs["lanelet_configs"]
        self.surrounding_configs = configs["surrounding_configs"]
        self.max_obs_dist: float = 0.0
        if self.surrounding_configs.get("observe_lane_circ_surrounding", False):
            self.max_obs_dist = self.surrounding_configs["lane_circ_sensor_range_radius"]
        elif self.surrounding_configs.get("observe_lane_rect_surrounding", False):
            self.max_obs_dist \
                = np.sqrt((self.surrounding_configs["lane_rect_sensor_range_length"] / 2) ** 2
                          + (self.surrounding_configs["lane_rect_sensor_range_width"] / 2) ** 2)
        elif self.surrounding_configs.get("observe_lidar_circle_surrounding", False):
            self.max_obs_dist = self.surrounding_configs["lidar_sensor_radius"]

        self.stopped_once = False

    def reset(self, observation_dict: dict, ego_action: Action):
        pass

    def calc_reward(self, observation_dict: dict, ego_action: Action) -> float:
        """
        Calculate the reward according to the observations

        :param observation_dict: current observations
        :param ego_action: Current ego_action of the environment
        :return: Reward of this step
        """
        reward = 0.0
        termination_reward = self.termination_reward(observation_dict)
        reward += termination_reward
        if termination_reward:
            # extra reward for being close the goal when terminating
            reward += self.long_distance_to_reference_path_reward(observation_dict)

        # Penalize reverse driving
        if self.reward_configs["reward_reverse_driving"]:
            reward += self.reverse_driving_penalty(ego_action)

        # Same lane as goal
        if self.reward_configs["reward_same_lane_goal"] > 0.:
            reward += self.goal_same_lane_reward(observation_dict)

        # Distance advancement
        if self.goal_configs["observe_distance_goal_long"] and self.goal_configs["observe_distance_goal_lat"]:
            reward += self.goal_distance_reward(observation_dict)

        # Closing in on goal-time
        if self.goal_configs["observe_distance_goal_time"]:
            reward += self.goal_time_reward(observation_dict, ego_action)
        if (
                (not self.goal_configs["observe_distance_goal_lat"]
                 or np.isclose(observation_dict["distance_goal_lat"][0], 0))
                and
                (not self.goal_configs["observe_distance_goal_long"]
                 or np.isclose(observation_dict["distance_goal_long"][0], 0))
                and
                (not self.goal_configs["observe_distance_goal_time"]
                 or observation_dict["distance_goal_time"] == 0)
        ):
            # Closing in on goal-orientation
            if self.goal_configs["observe_distance_goal_orientation"]:
                reward += self.goal_orientation_reward(observation_dict)

            # Closing in on goal-speed
            if self.goal_configs["observe_distance_goal_velocity"]:
                reward += self.goal_velocity_reward(observation_dict, ego_action)

        # Safe distance reward
        if self.reward_configs["reward_safe_distance_coef"]:
            reward += self.safe_distance_reward(observation_dict)

        # Traffic rule priority yield reward
        #      if self.reward_configs["yield_reward"]:
        #         reward += self.yield_reward(observation_dict)


        # orientation to reference path's orientation reward
        if self.lanelet_configs["observe_route_reference_path"]:
           reward+=self. orientation_to_reference_reward(observation_dict)

        # TODO: was commented out in original reward, needs to be reworked/removed?
        # penalize running given stop sign
        #    if self.traffic_sign_configs["observe_stop_sign"]:
        #       if self.stopped_once and observation_dict["stop_sign_distance_long"] == -1.0:
        #           self.stopped_once = False
        #       if observation_dict["stop_sign"][0]:
        #           reward += self.traffic_sign_reward(observation_dict)

        # TODO: was commented out in original reward, needs to be reworked/removed?
        # reward += self.deviation_lane_center_penalty()

        # TODO: was commented out in original reward, needs to be reworked/removed?
        # # passenger comfort
        # reward += self.passenger_comfort_reward()

        # TODO: was commented out in original reward, needs to be reworked/removed?
        # # penalize large lateral velocity using PM model
        # if self.ego_action.vehicle.vehicle_model == VehicleModel.PM:
        #     reward += self.large_lat_velocity_penalty()
        return reward

    def termination_reward(self, observation_dict: dict) -> float:
        """Reward for the cause of termination"""
        # Reach goal
        if observation_dict["is_goal_reached"][0]:
            LOGGER.debug("GOAL REACHED!")
            return self.reward_configs["reward_goal_reached"]
        # Collision
        if observation_dict["is_collision"][0]:
            return self.reward_configs["reward_collision"]
        # Off-road
        if observation_dict["is_off_road"][0]:
            return self.reward_configs["reward_off_road"]
        # Friction violation
        if observation_dict["is_friction_violation"][0]:
            return self.reward_configs["reward_friction_violation"]
        # Exceed maximum episode length
        if observation_dict["is_time_out"][0]:
            return self.reward_configs["reward_time_out"]

        return 0.0

    def goal_distance_reward(self, observation_dict: dict) -> float:
        """Reward for getting closer to goal distance"""
        long_advance = observation_dict["distance_goal_long_advance"][0]
        lat_advance = observation_dict["distance_goal_lat_advance"][0]

        return self.reward_configs["reward_closer_to_goal_long"] * long_advance + \
               self.reward_configs["reward_closer_to_goal_lat"] * lat_advance

    @staticmethod
    def _time_to_goal_weight(ego_velocity, distance, goal_time_distance):
        """
        Weight function for the goal_time_reward.
        """
        goal_time_distance = abs(goal_time_distance)
        if np.isclose(distance, 0):
            return 1.
        if not np.isclose(ego_velocity, 0):
            time_to_goal = abs(distance / ego_velocity)
            if not np.isclose(time_to_goal, 0):
                return min(time_to_goal / goal_time_distance, goal_time_distance / time_to_goal)
        return 0.

    def goal_same_lane_reward(self, observation_dict: dict):
        if observation_dict["distance_goal_lat"] < 2.:
            return self.reward_configs["reward_same_lane_goal"]
        else:
            return 0.

    def goal_time_reward(self, observation_dict: dict, ego_action: Action) -> float:
        """Reward for getting closer to goal time"""
        # TODO: improve time reward?
        # Idea: take distance to goal and velocity into account by approximating time to goal at current velocity
        if observation_dict["distance_goal_time"][0] >= 0:
            return 0.0
        elif self.goal_configs["observe_euclidean_distance"]:
            return self._time_to_goal_weight(
                ego_action.vehicle.state.velocity,
                observation_dict["euclidean_distance"][0], observation_dict["distance_goal_time"][0]) * \
                   self.reward_configs["reward_get_close_goal_time"]
        else:
            return self.reward_configs["reward_get_close_goal_time"]

    def goal_velocity_reward(self, observation_dict: dict, ego_action: Action) -> float:
        """Reward for getting closer to goal velocity"""
        if ego_action.vehicle.vehicle_model == VehicleModel.PM:
            velocity = np.sqrt(ego_action.vehicle.state.velocity ** 2 + ego_action.vehicle.state.velocity_y ** 2)
        else:
            velocity = abs(ego_action.vehicle.state.velocity)

        return self.reward_configs["reward_close_goal_velocity"] \
               * np.exp(-1.0 * abs(observation_dict["distance_goal_velocity"][0])
                        / (-1 * observation_dict["distance_goal_velocity"][0] + velocity))

    def goal_orientation_reward(self, observation_dict: dict) -> float:
        """
        Reward for getting closer to goal orientation
        """
        return self.reward_configs["reward_close_goal_orientation"] \
               * np.exp(-1.0 * abs(observation_dict["distance_goal_orientation"][0]) / np.pi)

    def reverse_driving_penalty(self, ego_action: Action) -> float:
        """Penalty for driving backwards"""
        return self.reward_configs["reward_reverse_driving"] * (ego_action.vehicle.state.velocity < 0.0)

    def safe_distance_reward(self, observation_dict: dict) -> float:
        """Reward for keeping a safe distance to the leading vehicle. If the ego vehicle is changing lanes, not keeping
        a safe distance to the following vehicle is also penalized"""
        # TODO: fix safe distance reward

        reward = 0.0
        a_max = 11.5 # TODO: load a_max from vehicle parameters
        lane_change = self.surrounding_configs["observe_lane_change"] and observation_dict["lane_change"] > 0.
        # dist_lead = self.max_obs_dist
        # dist_follow = self.max_obs_dist
        # v_rel = [v_rel_left_follow, v_rel_same_follow, v_rel_right_follow, v_rel_left_lead, v_rel_same_lead,
        #          v_rel_right_lead]
        # p_rel = [p_rel_left_follow, p_rel_same_follow, p_rel_right_follow, p_rel_left_lead, p_rel_same_lead,
        #          p_rel_right_lead]
        if self.surrounding_configs["observe_lane_rect_surrounding"] \
                or self.surrounding_configs["observe_lane_circ_surrounding"]:
            dist_lead = observation_dict["lane_based_p_rel"][4]
            dist_follow = observation_dict["lane_based_p_rel"][1]
            v_rel_lead = observation_dict["lane_based_v_rel"][4]
            v_rel_follow = observation_dict["lane_based_v_rel"][1]
            v_ego = np.sqrt(np.sum(observation_dict["v_ego"] ** 2))
            v_lead = v_rel_lead + v_ego
            v_follow = v_ego - v_rel_follow
            safe_dist_lead = max((v_ego ** 2 - v_lead ** 2) / (2 * a_max), 4.)
            safe_dist_follow = max((v_follow ** 2 - v_ego ** 2) / (2 * a_max), 4.)
        # elif self.surrounding_configs["observe_lidar_circle_surrounding"]:
        #     [dist_leading, dist_following] = observation_dict["dist_lead_follow_rel"]
        else:
            raise NotImplementedError(f"Safe distance reward is only supported for lane-based observations currently!")

        reward += self._safe_distance_reward_function(dist_lead, safe_dist_lead)
        if lane_change:
            reward += self._safe_distance_reward_function(dist_follow, safe_dist_follow)

        assert isinstance(reward, float)
        return reward

    def _safe_distance_reward_function(self, distance: float, safe_distance: float) -> float:
        """Exponential reward function for keeping a safe distance"""
        if distance < safe_distance:
            return self.reward_configs["reward_safe_distance_coef"] * np.exp(-5. * distance / safe_distance)
        else:
            return 0.

    def traffic_sign_reward(self, observation_dict: dict) -> float:
        """Reward for obeying traffic sings"""
        reward = 0.0
        #        dis_stop_sign = observation_dict["stop_sign_distance_long"][0]
        # if distance between ego vehicle and stop sign shorter than 2m
        #        if 0 < dis_stop_sign < 2 and observation_dict["v_ego"][0] > 0:
        #            # penalize large velocity
        #            reward += self.reward_configs['reward_stop_sign_vel'] * np.exp(
        #               -5 * dis_stop_sign / 2) * abs(observation_dict["v_ego"][0])
        #            # reward += self.reward_configs['reward_stop_sign_vel'] * (1 - 0.5 * dis_stop_sign) * \
        #            # observation_dict["v_ego"][0]
        #            # positive reward for negative acceleration(the ego vehicle should decelerate)
        #            if observation_dict["a_ego"][0] < 0:
        #                reward += self.reward_configs['reward_stop_sign_acc']
        #        # Finally, add constant reward for stopping in front of the stop sign once:
        #        if not self.stopped_once and np.isclose(observation_dict["v_ego"], 0.0):
        #            reward += self.reward_configs["stop_sign_vel_zero"]
        #            self.stopped_once = True
        return reward

    def deviation_lane_center_penalty(self, observation_dict: dict) -> float:
        """Penalty for deviating from the lane center"""
        # reward = 0.0
        # # Deviation from lane center
        # # TODO: use min. road edge distance instead
        # if self.lanelet_configs["observe_lat_offset"]:
        #     reward += -self.reward_configs["reward_stay_in_road_center"] * \
        #               np.exp(np.abs(self.observation_dict["lat_offset"][0]))
        #
        # if self.lanelet_configs["observe_left_road_edge_distance"] or self.lanelet_configs[
        #     "observe_right_road_edge_distance"]:
        #     min_road_dist = np.min((self.observation_dict["left_road_edge_distance"][0],
        #                             self.observation_dict["right_road_edge_distance"][0]))
        #     reward += -self.reward_configs["reward_stay_in_road_center"] * np.exp(-min_road_dist)
        # # Degree of violation of friction constraint
        # if (
        #         self.observe_a_ego
        #         and self.observe_v_ego
        #         and self.observe_steering_angle
        # ):
        #     a_ego = self.observation_dict["a_ego"][0]
        #     v_ego = self.observation_dict["v_ego"][0]
        #     steering_ego = self.observation_dict["steering_angle"][0]
        #     l_wb = self.ego_vehicle.params.a + ego_vehicle.params.b
        #     a_max = self.ego_vehicle.params.longitudinal.a_max
        #     reward += self.reward_friction * (
        #             a_max
        #             - (a_ego ** 2 + (v_ego ** 2 * np.tan(steering_ego) / l_wb) ** 2)
        #             ** 0.5
        #     )
        #
        # return reward

    def passenger_comfort_reward(self) -> float:
        """Penalty for bad passenger comfort"""
        # if self.current_step >= 2:
        #     ego_vehicle = self.ego_action.vehicle
        #     jerk_long = (ego_vehicle.state.acceleration - ego_vehicle.previous_state.acceleration) \
        #         / self.scenario.dt
        #     jerk_lat = (ego_vehicle.state.acceleration_y - ego_vehicle.previous_state.acceleration_y) \
        #         / self.scenario.dt
        #     return -1.0 * self.reward_configs["reward_jerk_long"] * np.abs(jerk_long) \
        #         + self.reward_configs["reward_jerk_lat"] * np.abs(jerk_lat)

    def large_lat_velocity_penalty(self) -> float:
        """Penalty for large lateral velocity using PM model"""
        # longitudinal and lateral velocity
        # v_long = self.ego_action.vehicle.state.velocity
        # v_lat = self.ego_action.vehicle.state.velocity_y
        # # print("current_long_v:", v_long, "current_lat_v", v_lat)
        # return self.reward_configs['reward_lateral_velocity'] * np.abs(v_lat)

    def lat_distance_to_reference_path_reward(self, observation_dict, distance50pctreward=3) -> float:
        """
        :param: distance50pctreward: distance in m for which 50% reward is returned 

        postive reward for staying close to the reference path in lateral direction
        max reward: 0.999 (0 m to the reference pathin lat direction)
        max penalty: - 0.0 (inf / >20 m to the reference pathin lat direction)
        """
        if "distance_togoal_via_referencepath" not in observation_dict:
            return 0.0
        long_lat_pos = observation_dict["distance_togoal_via_referencepath"]

        lat = abs(long_lat_pos[1])

        # sigmoid_distance ~= 1 for distance 0m to reference path, ~=0.5 for distance50pctreward m, ~0.01 for 2 * distance50pctreward m
        sigmoid_distance = 1 - 1 / (1 + exp(distance50pctreward - lat))

        return self.reward_configs['reward_lat_distance_reference_path'] * sigmoid_distance

    def long_distance_to_reference_path_reward(self, observation_dict, distance50pctreward=5) -> float:
        """
        reward/penalty for distance to goal on the reference path in longitudinal direction
        max reward: 2.0, for long is 0.
        max reward: 1.0, when being more then distance50pctreward meters before goal
        max penalty: - log( abs(long_distance_past_goal) + 1), when past the goal
        """
        if "distance_togoal_via_referencepath" not in observation_dict:
            return 0.0
        long_lat_pos = observation_dict["distance_togoal_via_referencepath"]

        long = long_lat_pos[0]

        if distance50pctreward > long >= 0:
            return 2.0 - (long / distance50pctreward)
        elif long > distance50pctreward:
            # if before goal
            # sigmoid_distance ~= 1 for distance 0m to reference path, ~=0.5 for 5m, ~0.01 for 10m
            return min(1 / (long * distance50pctreward), 1.0) * self.reward_configs[
                'reward_long_distance_reference_path']

        elif long < 0:
            # if driven past the goal: penalty
            return - np.log(abs(long) + 1) * self.reward_configs['reward_long_distance_reference_path']

    def orientation_to_reference_reward(self,observation_dict:dict) ->float:
        """
        Reward for getting closer to reference orientation
        ego orientation closer to the reference orientation, larger the reward
        """
        # see configs.yaml:  distances_route_reference_path, choose the index that are closest to ego vehicle
        ori_to_ref_diff = abs(observation_dict["route_reference_path_orientations"][1])
        return self.reward_configs['reward_orientation_to_ref']* np.exp(-1.0 * ori_to_ref_diff/ np.pi)



import argparse
import os
import re
import glob
import numpy as np
import yaml

import matplotlib.pyplot as plt
import imageio

from tqdm import tqdm

from SALib.sample import saltelli, fast_sampler, latin, morris, ff
from SALib.analyze import sobol, fast, rbd_fast, delta
from SALib.analyze import morris as morris_analyze
from SALib.analyze import ff as ff_analyze

from stable_baselines.common.vec_env import VecNormalize
from stable_baselines import PPO2
from stable_baselines.common import BaseRLModel
from stable_baselines.common.vec_env.util import copy_obs_dict, dict_to_obs, obs_space_info
from commonroad_rl.utils_run.utils import ALGOS, get_wrapper_class, make_env
from commonroad_rl.gym_commonroad.commonroad_env import CommonroadEnv
from commonroad_rl.gym_commonroad.constants import PATH_PARAMS

from commonroad_rl.gym_commonroad.utils.scenario_io import get_project_root

# TODO DGSM
# TODO ff doesnt work due to padding --> workaround?
os.environ["KMP_WARNINGS"] = "off" 

SAMPLER = {
    "sobol": saltelli,
    "fast": fast_sampler,
    "rbd_fast": latin,
    "morris": morris,
    "delta": latin,
    "ff": ff
}

ANALYZERS = {
    "sobol": sobol,
    "fast": fast,
    "rbd_fast": rbd_fast,
    "morris": morris_analyze,
    "delta": delta,
    "ff": ff_analyze
}

RESULTS = {
    "sobol": ['S1', 'S2', 'ST'],
    "fast": ['S1', 'ST'],
    "rbd_fast": ['S1'],
    "morris": ['mu', 'mu_star', 'sigma', 'mu_star_conf'],
    'delta': ['delta', 'delta_conf', 'S1', 'S1_conf'],
    "ff": ['ME', 'IE'],
    }


def get_parser():
    """
    Initializes Parser
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-n", help="Set the number of Samples to be considered", default=int(256), type=int)
    parser.add_argument("--algo", type=str, default="ppo2")
    parser.add_argument("--method", type=str, default=None, help="Sensitivity Analysis methods: sobol, fast, rbd_fast, morris")
    parser.add_argument("--config_filename", "-config_f", type=str, default="environment_configurations.yml", 
                        help="Name of the configuration file, default: environement_configurations.yml")
    parser.add_argument("--model_path", "-model", type=str, help="Path to trained model",
                        default=PATH_PARAMS["log"] + "/ppo2/commonroad-v1_1")
    parser.add_argument("--save_path", type=str, help="Path where the data should be saved",
                        default=PATH_PARAMS["log"])
    parser.add_argument("--data_path", type=str, help="Path to pickle files",
                        default=PATH_PARAMS["pickles"])
    parser.add_argument("--save_fig", action="store_true", help="Store plots for analysis")
    parser.add_argument("--save_data", action="store_true", help="Store data for analysis as .npy files")
    parser.add_argument("--save_gif", action="store_true", help="Store data for analysis as gif")
    parser.add_argument("--n_frames", type=int, default=5)

    return parser

def args_assertions(args):
    """
    This function collects assertions related to the parsed arguments

    :param args: the arguments input through the parser
    """

    assert (args.n & (args.n-1) == 0) and args.n != 0, '-n given as ' + str(args.n) + ' needs to be of 2^(x) with x a natural Number'
    assert args.method in ANALYZERS, '--method given as ' + args.method + ' is not in possible methods: sobol, fast, rbd_fast, morris'
    assert args.algo in ALGOS, '--alorithn given as ' + args.method + ' is not in possible alorithms: a2c, acer, acktr, ddpg, her, sac, pp2, trpo, td3'


def open_configs_sens_bounds():
    """
    This function loads the bounds from the config.yaml file in PATH_PARAMS["configs"]["commonroad-v1"]
    PATH_PARAMS are read from constants.py
    :return: a dict with the bound stuctured as: 'name_bound': [lower_b, upper_b]
    """
    config_file=PATH_PARAMS["configs"]["commonroad-v1"]
    with open(config_file, "r") as config_file:
            config = yaml.safe_load(config_file)

        # Assume default environment configurations
    return config["sensititvity_analysis_bounds"]


def load_model(model_path: str, algo: str) -> BaseRLModel:
    """
    Load trained model

    :param model_path: Path to the trained model
    :param algo: The used RL algorithm
    """
    # Load the trained agent
    # TODO: load last model if best_model.zip does not exist (no evaluation env was created during training)
    files = os.listdir(model_path)
    if "best_model.zip" in files:
        model_path = os.path.join(model_path, "best_model.zip")
    else:
        # find last model
        files = sorted(glob.glob(os.path.join(model_path, "rl_model*.zip")))
        def extract_number(f):
            s = re.findall("\d+", f)
            return (int(s[-1]) if s else -1, f)
        model_path = max(files, key=extract_number)
    model = [ALGOS[algo].load(model_path)]

    return model


def load_all_models(model_path: str, algo: str) -> BaseRLModel:
    """
    Load trained model

    :param model_path: Path to the trained model
    :param algo: The used RL algorithm
    """
    # Load all Models
    def extract_number(f):
            s = re.findall("\d+", f)
            return (int(s[-1]) if s else -1, f)
        
    files = sorted(glob.glob(os.path.join(model_path, "rl_model*.zip")), key=extract_number)
    models = []
    for each in files:
        print(f"loading model {each}")
        models.append(ALGOS[algo].load(each))
    return models


def load_env(args):
    """
    This function loads an enviromenet based on the env_configs

    :param args: the arguments input through the parser, needed for model_path, data_path
    :return: a Commonroad Env
    """

    #Load Kwargs
    env_configs = {}
    with open(os.path.join(args.model_path, "environment_configurations.yml"), "r") as config_file:
        env_configs = yaml.safe_load(config_file)
    env_kwargs = env_configs  

    # Create environment
    # note that CommonRoadVecEnv is inherited from DummyVecEnv
    env_kwargs['test_reset_config_path'] = args.data_path + '//problem_test'
    env_kwargs['train_reset_config_path'] = args.data_path + '//problem_train'
    env_kwargs['meta_scenario_path'] = args.data_path + '//meta_scenario'
    env = CommonroadEnv(**env_kwargs)
    return env

def helper_compress_figure_labelx(labels, values):
    """
    This function makes the plot smaller (compresses the x axis of the plot) by taking the mean abs. value of multiple same type observations

    :param labels: list of labels, np.ndarray or 1D List of strings
    :param values: list of values, np.ndarray
    :return: Tuple(labels_new, values_new)
    """
    assert len(labels) == len(values), f"values: {values} and labels: {labels} need to be same lenght"
    # find unique labels and indices
    indexdict = {}
    for index, name_l in enumerate(labels):
        found_similar = False
        for exis in indexdict.keys():
            if name_l.startswith(exis):
                indexdict[exis].append(index) 
                found_similar=True
                break
        if not found_similar:
            indexdict[name_l]= [index]
    
    # new values
    values_new = []

    for indices in indexdict.values():
        values_new.append(np.average(np.abs(values[indices])))
    
    return list(indexdict.keys()), np.array(values_new)

def refactor_name_list(name_list):
    """
    This function removes the numbers of the observation names if the observation is a singular observation

    :param name_list: list of the names associated with the observations like ['[obs1]_00', [obs2]_00', '[obs2]_01', ...]
    """
    for i, each in enumerate(name_list):
        if each[-1] == '0' and each[-2] == '0':
            name_list[i] = each[:-3]
    return name_list


def plot_results_1dim(labels, values, meaning, path, ylim=(-0.1, 0.4), fformat="png"):
    """
    This function plots 1 dimensional resutls of a sensitivity analysis as a bar chart.
    This plot is saved at the in args specified location
    :param labels: the names associated with the individual observations as a list
    :param values: the values associated with the individual observations as a list in the same order as the labels
    :param meaning: a string describing the meaning of the plot
    :param path: save path
    """

    x_pos = [i for i, _ in enumerate(labels)]
    breite = len(labels) * 0.17
    plt.subplots(figsize=(breite, 8))
    plt.gcf().subplots_adjust(bottom=0.60)

    plt.bar(x_pos, values, color='green')

    plt.xlabel("Observations")
    plt.ylabel("Sensitivity value for each Obs")
    plt.title("Sensitivity analysis " + meaning)
    plt.xticks(x_pos, labels, rotation=90)

    plt.ylim(1.1 * ylim[0], 1.1 * ylim[1])


    plt.savefig(path + '/sens_analysis_' + meaning + '.' + fformat)
    plt.close()


def plot_results_2dim(labels, values, meaning, path, fformat="png"):
    """
    This function plots 2 dimensional resutls of a sensitivity analysis as a bar chart.
    This plot is saved at the in args specified location
    :param labels: the names associated with the individual observations as a list
    :param values: the values associated with the individual observations as a list in the same order as the labels
    :param meaning: a string describing the meaning of the plot
    :param path: save path    
    """

    for each in labels:
        if each[-1]  == '0' and each[-2] == '0':
            each = each[0:-3]  

    fig, ax = plt.subplots(figsize=(30, 30))
    im = ax.imshow(values)

    # We want to show all ticks...
    ax.set_xticks(np.arange(len(labels)))
    ax.set_yticks(np.arange(len(labels)))
    # ... and label them with the respective list entries
    ax.set_xticklabels(labels)
    ax.set_yticklabels(labels)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=90, ha="right",
            rotation_mode="anchor")

    # Loop over data dimensions and create text annotations.
    for i in range(len(labels)):
        for j in range(len(labels)):
            text = ax.text(j, i, round(values[i, j],2),
                        ha="center", va="center", color="w")

    ax.set_title("Sensitivity analysis " + meaning)
    fig.tight_layout()
    plt.savefig(path + '/sens_analysis_' + meaning + '.' + fformat)
    plt.close()



def perform_analysis(args):
    """
    Performs the whole analysis workflow
    If args.save_fig is true, the plots are saved
    The single steps are given in inline comments
    :param args: :param args: parsed input arguments
    """
    # Load environement
    env = load_env(args)
    # Load pretrained model
    if args.save_gif == False:
        models = load_model(args.model_path, args.algo)
    else:
        models = load_all_models(args.model_path, args.algo)


    
    # add observation w observation bounds
    env.reset()
    obs = env.observation_dict
    num_vars = 0
    name_list =[]
    for key in obs:       
        num_vars += len(obs[str(key)])
        for i in range(len(obs[str(key)])):
            name_list.append(str(key + '_' + "{0:02}".format(i)))



    # get sample bounds
    bound_list = []
    bound_configs = open_configs_sens_bounds()
    for each in name_list:
        assert each[:-3] in bound_configs, str(each[:-3]) \
            + ' missing bound, please add in respective section of configs (on the bottom)' 
        bound_list.append(bound_configs[each[:-3]])

    problem = {
    'num_vars': num_vars,
    'names': name_list,
    'bounds': bound_list
    }

    # generate Samples
    param_values = SAMPLER[args.method].sample(problem, args.n)  

    
    # correct sampling errrors with boolean bounds
    for sample in param_values:
        for i in range(len(bound_list)):
            if True in bound_list[i]:
                sample[i] = np.random.choice(bound_list[i])
    
    if args.save_gif:
        pic_dict1 = {}
        pic_dict2 = {}
        for each in RESULTS[args.method]: 
                    pic_dict1[each] = []
                    pic_dict2[each] = []   
    values1 = []
    values2 = []

    for model_num, model in tqdm(enumerate(models), desc="Evaluation all Models"):
        print(f"performing analysis on model {model_num}  / {len(models)}")
        # for n defined in args collect result of model.predict
        Model_predictions = []
        for sample in param_values:
            y = model.predict(sample, deterministic=True)
            Model_predictions.append(y[0])
        Model_predictions = np.array(Model_predictions, dtype=float)
        Model_predictions = np.transpose(Model_predictions)
        Y = np.zeros([param_values.shape[0]])
        
        # Perform analysis
        if args.method == 'rbd_fast' or args.method == 'morris' or args.method == 'delta' or args.method == 'ff':
            Si = ANALYZERS[args.method].analyze(problem, param_values, Model_predictions[0], print_to_console=False)
            Si2 = ANALYZERS[args.method].analyze(problem,param_values, Model_predictions[1], print_to_console=False)
        else:
            Si = ANALYZERS[args.method].analyze(problem, Model_predictions[0], print_to_console=False)
            Si2 = ANALYZERS[args.method].analyze(problem, Model_predictions[1], print_to_console=False)

        # Output analysis results
        # Plot result

        name_list = refactor_name_list(name_list)

        if args.save_data and model == models[-1]:
            path = os.path.join(args.save_path, 'sens_analysis_data')
            
            os.makedirs(path, exist_ok=True)
            saved_arrays = {
                f"{args.method}_labels": np.asarray(name_list),
            }
            for k, v in Si.items():
                saved_arrays[f"{args.method}_action1_{k}"] = v
            for k, v in Si.items():
                saved_arrays[f"{args.method}_action2_{k}"] = v
            
            np.savez(os.path.join(path, f"sens_analysis_{args.method}"), **saved_arrays)

        if args.save_fig and model == models[-1]:
            print("saving figs")
            path = os.path.join(args.save_path, 'sens_analysis_figs')
            os.makedirs(path, exist_ok=True)

            for each in RESULTS[args.method]:
                meaning1 = str(args.method + "_action1_" + each)
                meaning2 = str(args.method + "_action2_" + each)  
                if each == 'S2' or each == 'IE':
                    plot_results_2dim(name_list, Si2[each], meaning2, path)
                    plot_results_2dim(name_list, Si[each], meaning1, path)
                else:
                    plot_results_1dim(name_list, Si[each], meaning1, path, (min(Si[each]), max(Si[each])))
                    plot_results_1dim(name_list, Si2[each], meaning2, path, (min(Si2[each]), max(Si2[each])))
        
        if args.save_gif:
            values1.append(Si)
            values2.append(Si2)

    if args.save_gif:
        path = os.path.join(args.save_path, 'sens_analysis_figs')

        for each in tqdm(RESULTS[args.method], desc="Constructing gifs"):
            max_value = max(max(value[each]) for value in values1)
            min_value = min(min(value[each]) for value in values1)
            for val_iter in tqdm(range(0, len(values1) - 1), desc=f"-Images for first action of {each}"):
                results = values1[val_iter][each]
                next_results = values1[val_iter + 1][each]
                dist_results = np.array(next_results) - np.array(results)
                for i in range(0, args.n_frames + 1):
                    interpol_results = (results + (dist_results/args.n_frames) * i)
                    meaning = str(args.method + "_action1_" + each) + "_" + str(val_iter) + "_" + str(i)
                    pic_dict1[each].append(path + '/sens_analysis_' + meaning + '.png')
                    if i == args.n_frames:
                        for j in range(0, 5):
                            pic_dict1[each].append(path + '/sens_analysis_' + meaning + '.png')
                    if each == 'S2' or each == 'IE':
                        plot_results_2dim(name_list, interpol_results, meaning, path)
                    else:
                        plot_results_1dim(name_list, interpol_results, meaning, path, ylim=(min_value, max_value))

            max_value = max(max(value[each]) for value in values2)
            min_value = min(min(value[each]) for value in values2)
            for val_iter in tqdm(range(0, len(values2) - 1), desc=f"-Images for second action {each}"):
                results = values2[val_iter][each]
                next_results = values2[val_iter + 1][each]
                dist_results = np.array(next_results) - np.array(results)
                for i in range(0, args.n_frames + 1):
                    interpol_results = (results + (dist_results/args.n_frames) * i)
                    meaning = str(args.method + "_action2_" + each) + "_" + str(val_iter) + "_" + str(i)
                    pic_dict2[each].append(path + '/sens_analysis_' + meaning + '.png')
                    if i == args.n_frames:
                        for j in range(0, 5):
                            pic_dict2[each].append(path + '/sens_analysis_' + meaning + '.png')
                    if each == 'S2' or each == 'IE':
                        plot_results_2dim(name_list, interpol_results, meaning, path)
                    else:
                        plot_results_1dim(name_list, interpol_results, meaning, path, ylim=(min_value, max_value))
            
            new_path = os.path.join(path, 'gifs')
            os.makedirs(path, exist_ok=True)

            gif_path  = new_path + "/" + each + '_action1.gif'           
            with imageio.get_writer(gif_path, mode = 'I') as writer:
                for filename in tqdm(pic_dict1[each], desc=f"-Gif for first action {each}"):
                    image = imageio.imread(filename)
                    writer.append_data(image)

            for filename in set(pic_dict1[each]):
                os.remove(filename)

            gif_path  = new_path + "/" + each + '_action2.gif'
            with imageio.get_writer(gif_path, mode = 'I') as writer:
                for filename in tqdm(pic_dict2[each], desc=f"-Gif for second action {each}"):
                    image = imageio.imread(filename)
                    writer.append_data(image)

            for filename in set(pic_dict2[each]):
                os.remove(filename)



def main():
    """
    Main method, calling the parser and the sensititvity analysis
    """
    args = get_parser().parse_args()
    args_assertions(args)
    perform_analysis(args)

if __name__ == "__main__":
    main()



"""Class for dense reward"""
import warnings

import numpy as np

from commonroad_rl.gym_commonroad.action import Action
from commonroad_rl.gym_commonroad.reward.reward import Reward


class DenseReward(Reward):
    """Class for dense reward"""

    def __init__(self, configs: dict):
        self.reward_configs = configs["reward_configs_dense"]

        self.surrounding_configs = configs["surrounding_configs"]
        self.max_obs_dist: float = 0.0
        if self.surrounding_configs["observe_lane_circ_surrounding"]:
            self.max_obs_dist = self.surrounding_configs["lane_circ_sensor_range_radius"]
        elif self.surrounding_configs["observe_lane_rect_surrounding"]:
            self.max_obs_dist \
                = np.sqrt((self.surrounding_configs["lane_rect_sensor_range_length"] / 2) ** 2
                          + (self.surrounding_configs["lane_rect_sensor_range_width"] / 2) ** 2)
        elif self.surrounding_configs["observe_lidar_circle_surrounding"]:
            self.max_obs_dist = self.surrounding_configs["lidar_sensor_radius"]

        self.initial_goal_dist: float = -1.0

    def reset(self, observation_dict: dict, ego_action: Action):
        distance_goal_long = observation_dict["distance_goal_long"][0]
        distance_goal_lat = observation_dict["distance_goal_lat"][0]
        self.initial_goal_dist = np.sqrt(distance_goal_long ** 2 + distance_goal_lat ** 2)

        # Prevent cases where the ego vehicle starts in the goal region
        if self.initial_goal_dist < 1.0:
            warnings.warn("Ego vehicle starts in the goal region")
            self.initial_goal_dist = 1.0

    def calc_reward(self, observation_dict: dict, ego_action: Action) -> float:
        """
        Calculate the reward according to the observations

        :param observation_dict: current observations
        :param ego_action: Current ego_action of the environment
        :return: Reward of this step
        """

        # Calculate normalized distance to obstacles as a positive reward
        # Lane-based
        rel_pos = []
        if self.surrounding_configs["observe_lane_rect_surrounding"] or \
                self.surrounding_configs["observe_lane_circ_surrounding"]:
            rel_pos += observation_dict["lane_based_p_rel"].tolist()

        if len(rel_pos) == 0:
            r_obs_lane = 0.0
        else:
            # Minus 5 meters from each of the lane-based relative positions,
            # to get approximately minimal distances between vehicles,
            # instead of exact distances between centers of vehicles
            r_obs_lane = (self.reward_configs["reward_obs_distance_coefficient"] * (
                    np.sum(rel_pos) - 5.0 * len(rel_pos)) / (self.max_obs_dist * len(rel_pos)))

        # Lidar-based
        dist = []
        if self.surrounding_configs["observe_lidar_circle_surrounding"]:
            dist += observation_dict["lidar_circle_dist"].tolist()

        if len(dist) == 0:
            r_obs_lidar = 0.0
        else:
            r_obs_lidar = (self.reward_configs["reward_obs_distance_coefficient"] * np.sum(dist) / (
                    self.max_obs_dist * len(dist)))

        # Calculate normalized distance to goal as a negative reward
        dist_goal = np.sqrt(observation_dict["distance_goal_long"][0] ** 2
                            + observation_dict["distance_goal_lat"][0] ** 2)
        r_goal = -self.reward_configs["reward_goal_distance_coefficient"] * dist_goal / self.initial_goal_dist

        return r_obs_lane + r_obs_lidar + r_goal



#! /usr/bin/env python

__author__ = "Niels Mündler"
__copyright__ = ""
__credits__ = [""]
__version__ = "1.0"
__maintainer__ = "Xiao Wang"
__email__ = "xiao.wang@tum.de"
__status__ = "Released"

__desc__ = """
Validates a Commonroad scenario syntactically
"""

from argparse import ArgumentParser
from pathlib import Path

from commonroad.common.file_reader import CommonRoadFileReader
from commonroad.scenario.lanelet import LaneletNetwork
from commonroad.scenario.scenario import Scenario
from commonroad.scenario.obstacle import DynamicObstacle, StaticObstacle
from commonroad.planning.planning_problem import PlanningProblem

from commonroad_route_planner.route_planner import RoutePlanner

from lxml import etree
from shapely.validation import explain_validity
from sys import exit


def check_adjacency(lanelet_network: LaneletNetwork):
    """
    Check that adjacencies are assigned correctly and bilaterally
    """
    errors = 0
    for l in lanelet_network.lanelets:
        if l.adj_left is not None:
            l_left = lanelet_network.find_lanelet_by_id(l.adj_left)
            if l.adj_left_same_direction and l_left.adj_right != l.lanelet_id:
                print(
                    f"Left of lanelet {l.lanelet_id} is {l.adj_left} facing the same direction but right of {l_left.lanelet_id} is {l_left.adj_right}"
                )
                errors += 1
            if not l.adj_left_same_direction and l_left.adj_left != l.lanelet_id:
                print(
                    f"Left of lanelet {l.lanelet_id} is {l.adj_left} facing the opposite direction but left of {l_left.lanelet_id} is {l_left.adj_left}"
                )
                errors += 1
        if l.adj_right is not None:
            l_right = lanelet_network.find_lanelet_by_id(l.adj_right)
            if l.adj_right_same_direction and l_right.adj_left != l.lanelet_id:
                print(
                    f"Right of lanelet {l.lanelet_id} is {l.adj_right} facing the same direction but left of {l_right.lanelet_id} is {l_right.adj_left}"
                )
                errors += 1
            if not l.adj_right_same_direction and l_right.adj_right != l.lanelet_id:
                print(
                    f"Left of lanelet {l.lanelet_id} is {l.adj_right} facing the opposite direction but left of {l_right.lanelet_id} is {l_right.adj_left}"
                )
                errors += 1
    return errors


def check_valid_lanelet_polygon(lanelet_network: LaneletNetwork):
    errors = 0
    for lanelet in lanelet_network.lanelets:
        try:
            assert lanelet.convert_to_polygon().shapely_object.is_valid
        except AssertionError as e:
            print(f"Lanelet {lanelet.lanelet_id} has invalid geometry")
            print(explain_validity(lanelet.convert_to_polygon().shapely_object))
            errors += 1
    return errors


def check_successors(lanelet_network: LaneletNetwork):
    """
    Check that adjacencies are assigned correctly and bilaterally
    """
    errors = 0
    for l in lanelet_network.lanelets:
        if l.successor is not None:
            for i in l.successor:
                suc = lanelet_network.find_lanelet_by_id(i)
                if not suc.predecessor or not l.lanelet_id in suc.predecessor:
                    print(
                        f"Lanelet {i} is successor of {l.lanelet_id} but does not have {l.lanelet_id} as predecessor"
                    )
                    errors += 1
        if l.predecessor is not None:
            for i in l.predecessor:
                pred = lanelet_network.find_lanelet_by_id(i)
                if not pred.successor or not l.lanelet_id in pred.successor:
                    print(
                        f"Lanelet {i} is predecessor of {l.lanelet_id} but does not have {l.lanelet_id} as successor"
                    )
                    errors += 1
    return errors


def check_obstacle_off_road(scenario: Scenario):
    errors = 0
    for o in scenario.obstacles:
        off_road = True
        if isinstance(o, DynamicObstacle):
            for s in o.prediction.trajectory.state_list:
                if scenario.lanelet_network.find_lanelet_by_position([s.position]):
                    off_road = False
                    break
        elif isinstance(o, StaticObstacle):
            off_road = not scenario.lanelet_network.find_lanelet_by_position(
                [o.initial_state.position]
            )
        try:
            assert not off_road
        except AssertionError:
            print(f"Obstacle {o.obstacle_id} is off the road at all times")
            errors += 1
    return errors


def check_path_to_goal(scenario: Scenario, planning_problem: PlanningProblem):
    errors = 0
    route_planner = RoutePlanner(
        scenario,
        planning_problem,
        backend=RoutePlanner.Backend.NETWORKX_REVERSED,
        log_to_console=False,
    )
    if len(route_planner.get_route_candidates().route_candidates) == 0:
        # no path found
        print(f"No path from initial position to goal region")
        errors += 1
    return errors


def validate(xml_path: Path, xsd_path: Path) -> 0:
    # test whether a scenario file passes the schema test
    try:
        assert xml_path.exists()
        etree.clear_error_log()
        xmlschema = etree.XMLSchema(etree.parse(str(xsd_path)))
        tmp = etree.parse(str(xml_path))
        xmlschema.assertValid(tmp)
    except AssertionError:
        print("File not found")
        return 1
    except etree.DocumentInvalid as e:
        print("File invalid: {}".format(e))
        return 1
    except Exception as e:
        print("Unknown error: {}".format(e))
        return 1
    return 0


def get_parser():
    parser = ArgumentParser(description=__desc__)
    parser.add_argument(
        "input", help="Path to input file(s) (.xml commonroad scenario)", nargs="+"
    )
    parser.add_argument(
        "-s",
        help="Path to schema file (.xsd commonroad specification)",
        default="./XML_commonRoad_XSD_2020a.xsd",
        dest="spec",
    )
    return parser


def main():
    args = get_parser().parse_args()

    xsd_path = Path(args.spec)
    assert xsd_path.exists(), f"file path {xsd_path} doesn't exist"

    total_errors = 0
    for input_file in args.input:
        xml_path = Path(input_file)
        print(xml_path)
        errors = 0

        # Check XML validity
        errors += validate(xml_path, xsd_path)

        # Additionally check if adjacent right/left is always assigned respectively
        scenario, planning_problem_set = CommonRoadFileReader(str(xml_path)).open()
        lanelet_network = scenario.lanelet_network
        errors += check_adjacency(lanelet_network)

        # check for invalid lanelet polygons like self-intersections or similar
        errors += check_valid_lanelet_polygon(scenario.lanelet_network)

        # Check that no obstacle is off the road all the time
        errors += check_obstacle_off_road(scenario)

        # # Check that there exists a route from the initial position
        # # to the goal
        # for pp in planning_problem_set.planning_problem_dict.values():
        #     errors += check_path_to_goal(scenario, pp)

        total_errors += errors

    if total_errors > 0:
        print(f"{total_errors} Errors where detected")
        exit(-1)
    else:
        print("No errors")
        exit(0)


if __name__ == "__main__":
    main()



__author__ = "Brian Liao"
__copyright__ = "TUM Cyber-Physical System Group"
__credits__ = []
__version__ = "0.1"
__maintainer__ = "Brian Liao"
__email__ = "brian.liao@tum.de"
__status__ = "Integration"

"""
Unit test script for model hyperparameter optimizing
"""
from commonroad_rl.train_model import *
from commonroad_rl.tests.common.marker import *
from commonroad_rl.tests.common.path import *

resource_path = resource_root("test_gym_commonroad/pickles")
output_path = output_root("test_utils_run")

@pytest.mark.parametrize(("env_id"), [("commonroad-v1")])

@slow
@module_test
@functional
def test_optimize_hyperparams(env_id):

    # Run training
    meta_scenario_path = os.path.join(resource_path, "meta_scenario")
    train_reset_config_path = os.path.join(resource_path, "problem")
    test_reset_config_path = os.path.join(resource_path, "problem")
    visualization_path = os.path.join(output_path, "images")
    log_path = os.path.join(output_path, "log", env_id)

    goal_relaxation = False
    num_of_steps = 1000
    algo = "ppo2"
    # clean output folder before
    # os.path.join(log_path, algo)

    sampler = "tpe"
    pruner = "median"

    args_str = (
        f"--seed 5 "
        f"--algo {algo} "
        f"--optimize-hyperparams "
        f"--n-trials 10 "
        f"--n-jobs 1 "
        f"--n-timesteps {num_of_steps} "
        f"--eval-freq 500 "
        f"--eval-episodes 3 "
        f"--env {env_id} "
        f"--log-folder {log_path} "
        f"--sampler {sampler} "
        f"--pruner {pruner}"
        f" --env-kwargs"
        f' reward_type:"hybrid_reward"'
        f' meta_scenario_path:"{meta_scenario_path}"'
        f' train_reset_config_path:"{train_reset_config_path}"'
        f' test_reset_config_path:"{test_reset_config_path}"'
        f' visualization_path:"{visualization_path}"'
    )

    args = run_stable_baselines_argsparser().parse_args(args_str.split(sep=" "))
    run_stable_baselines(args)

    model_version = sorted(os.listdir(os.path.join(log_path, algo)))[-1]
    file_to_check = f"model_hyperparameter_optimization/" \
                    f"report_{algo}_{env_id}-10-trials-{num_of_steps}-steps-{sampler}-{pruner}.yml"
    file_path_to_check = os.path.join(log_path, algo, model_version, file_to_check)

    assert os.path.isfile(file_path_to_check)
